# This file is part of LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Sickle cell anemia vs. malaria, a classic example of population genetics.

This script will initialize LiSEworld.db and the game code libraries to run the
simulation described. To view it, run ELiDE from the same directory
as you ran this script from.

"""


import networkx as nx
from LiSE import Engine
from os import remove


def install(
        engine,
        n_creatures=5,
        n_sickles=3,
        malaria_chance=.05,
        mate_chance=.05,
        mapsize=(1, 1),
        startpos=(0, 0)
):
    """Natural Selection on Sickle Cell Anemia

    If anyone carries a pair of sickle betaglobin genes, they die of
    sickle cell anemia.

    Individuals with 1x betaglobin, 1x sickle betaglobin are immune to
    malaria.

    """
    initmap = nx.grid_2d_graph(*mapsize)
    phys = engine.new_character("physical", data=initmap)
    species = engine.new_character(
        "species",
        mate_chance=mate_chance,
        malaria_chance=malaria_chance,
        n_creatures=n_creatures,
    )
    for n in range(0, n_creatures):
        name = "critter" + str(n)
        phys.add_thing(
            name=name,
            location=startpos,
            sickle_a=(n < n_sickles),
            sickle_b=False,
            male=engine.coinflip(),
            last_mate_turn=-1
        )
        assert name in phys.thing
        assert name not in phys.place
        assert name in phys.node, "couldn't add node {} to phys.node".format(name)
        assert hasattr(phys.node[name], 'location')
        species.add_avatar("physical", name)
        assert hasattr(species.avatar['physical'][name], 'location')

    # putting dieoff earlier in the code than mate means that dieoff will
    # be followed before mate is
    @species.avatar.rule
    def dieoff(critter):
        critter.delete()
        assert (critter.name not in critter.character.node)
        if critter['from_malaria']:
            return 'malaria'
        else:
            return 'anemia'

    @species.avatar.rule
    def mate(critter):
        """If I share my location with another critter, attempt to mate"""
        suitors = list(
            oc for oc in critter.location.contents()
            if oc['male'] != critter['male']
        )
        assert (len(suitors) > 0)
        other_critter = critter.engine.choice(suitors)
        sickles = [
            critter['sickle_a'],
            critter['sickle_b'],
            other_critter['sickle_a'],
            other_critter['sickle_b']
        ]
        engine.shuffle(sickles)
        name = "critter" + str(species.stat["n_creatures"])
        species.stat["n_creatures"] += 1
        engine.character["physical"].add_thing(
            name,
            critter["location"],
            sickle_a=sickles.pop(),
            sickle_b=sickles.pop(),
            male=engine.coinflip(),
            last_mate_turn=engine.turn
        )
        species.add_avatar("physical", name)
        critter['last_mate_turn'] = other_critter['last_mate_turn'] = \
            engine.turn
        return 'mated'

    @mate.prereq
    def once_per_turn(critter):
        return critter['last_mate_turn'] < critter.engine.turn

    @mate.prereq
    def mate_present(critter):
        for oc in critter.location.contents():
            if oc['male'] != critter['male']:
                return True
        return False

    @mate.trigger
    def in_the_mood(critter):
        return critter.engine.random() < critter.user.stat['mate_chance']

    @dieoff.trigger
    def sickle2(critter):
        r = critter['sickle_a'] and critter['sickle_b']
        if r:
            critter['from_malaria'] = False
        return r

    @dieoff.trigger
    def malaria(critter):
        r = (
                critter.engine.random() < critter.user.stat['malaria_chance'] and not
        (critter['sickle_a'] or critter['sickle_b'])
        )
        if r:
            critter['from_malaria'] = True
        return r

    # it would make more sense to keep using species.avatar.rule, this
    # is just a test
    @phys.thing.rule
    def wander(critter):
        dests = list(critter.character.place.keys())
        dests.remove(critter['location'])
        dest = critter.engine.choice(dests)
        critter.travel_to(dest)

    @wander.trigger
    def not_travelling(critter):
        return critter.next_location is None

    @wander.prereq
    def big_map(critter):
        return len(critter.character.place) > 1


def sickle_cell_test(
        engine,
        n_creatures=5,
        n_sickles=3,
        malaria_chance=.05,
        mate_chance=.05,
        mapsize=(1, 1),
        startpos=(0, 0),
        turns=100
):
    install(engine, n_creatures, n_sickles, malaria_chance, mate_chance, mapsize, startpos)
    species = engine.character['species']
    print(
        "Starting with {} creatures, of which {} have "
        "at least one sickle betaglobin.".format(
            len(species.avatar['physical']),
            sum(
                1 for critter in species.avatar['physical'].values()
                if critter['sickle_a'] or critter['sickle_b']
            )
        )
    )

    for i in range(0, turns):
        malaria_dead = 0
        anemia_dead = 0
        born = 0
        while engine.turn < i:
            r = engine.next_turn()
            if not r:
                continue
            r = r[0]
            if isinstance(r, Exception):
                raise r
            if 'malaria' in r:
                malaria_dead += 1
            if 'anemia' in r:
                anemia_dead += 1
            if 'mated' in r:
                born += 1
        print("On turn {}, {} critters were born; "
              "{} died of malaria, and {} of sickle cell anemia, "
              "leaving {} alive.".format(
                  i, born, malaria_dead, anemia_dead,
                  len(engine.character['species'].avatar['physical'])
              ))
    print(
        "Of the remaining {} creatures, {} have a sickle betaglobin.".format(
            len(species.avatar['physical']),
            sum(
                1 for critter in species.avatar['physical'].values()
                if critter['sickle_a'] or critter['sickle_b']
            )
        )
    )


if __name__ == '__main__':
    try:
        remove('LiSEworld.db')
    except OSError:
        pass
    with Engine('LiSEworld.db', random_seed=69105) as engine:
        sickle_cell_test(engine)



# This file is part of LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""The sort of node that is ultimately located in a Place.

Things may be located in other Things as well, but eventually must be
recursively located in a Place.

There's a subtle distinction between "location" and "containment": a
Thing may be contained by a Portal, but cannot be located there --
only in one of the Portal's endpoints. Things are both located in and
contained by Places, or possibly other Things.

"""
import networkx as nx
from .node import Node
from .exc import TravelException
from allegedb.cache import HistoryError


def roerror(*args, **kwargs):
    raise ValueError("Read-only")


class Thing(Node):

    """The sort of item that has a particular location at any given time.

    If a Thing is in a Place, it is standing still. If it is in a
    Portal, it is moving through that Portal however fast it must in
    order to arrive at the other end when it is scheduled to. If it is
    in another Thing, then it is wherever that is, and moving the
    same.

    """
    __slots__ = ('graph', 'db', 'node')

    extrakeys = {
        'name',
        'character',
        'location'
    }

    def _getname(self):
        return self.name

    def _getcharname(self):
        return self.character.name

    def _getloc(self):
        return self.engine._things_cache.retrieve(
            self.character.name, self.name, *self.engine.btt()
        )

    def _get_arrival_time(self):
        charn = self.character.name
        n = self.name
        thingcache = self.engine._things_cache
        for b, trn, tck in self.engine._iter_parent_btt():
            try:
                v = thingcache.turn_before(charn, n, b, trn)
            except KeyError:
                v = thingcache.turn_after(charn, n, b, trn)
            if v is not None:
                return v
        else:
            raise ValueError("Couldn't find arrival time")

    def _set_loc(self, loc):
        self.engine._set_thing_loc(
            self.character.name,
            self.name,
            loc
        )
        self.send(self, key='location', val=loc)

    _getitem_dispatch = {
        'name': _getname,
        'character': _getcharname,
        'location': _getloc
    }

    _setitem_dispatch = {
        'name': roerror,
        'character': roerror,
        'arrival_time': roerror,
        'next_arrival_time': roerror,
        'location': _set_loc
    }

    def __contains__(self, key):
        if key in self.extrakeys:
            return True
        return super().__contains__(key)

    def __getitem__(self, key):
        """Return one of my stats stored in the database, or a few
        special cases:

        ``name``: return the name that uniquely identifies me within
        my Character

        ``character``: return the name of my character

        ``location``: return the name of my location

        """
        try:
            return self._getitem_dispatch[key](self)
        except KeyError:
            return super().__getitem__(key)

    def __setitem__(self, key, value):
        """Set ``key``=``value`` for the present game-time."""
        try:
            self._setitem_dispatch[key](self, value)
        except HistoryError as ex:
            raise ex
        except KeyError:
            super().__setitem__(key, value)

    def __delitem__(self, key):
        """As of now, this key isn't mine."""
        if key in self.extrakeys:
            raise ValueError("Can't delete {}".format(key))
        super().__delitem__(key)

    def __repr__(self):
        return "{}.character['{}'].thing['{}']".format(
            self.engine,
            self.character.name,
            self.name
        )

    def delete(self):
        super().delete()
        self._set_loc(None)
        self.character.thing.send(self.character.thing, key=self.name, val=None)

    def clear(self):
        """Unset everything."""
        for k in list(self.keys()):
            if k not in self.extrakeys:
                del self[k]

    @property
    def location(self):
        """The ``Thing`` or ``Place`` I'm in."""
        return self.engine._get_node(self.character, self['location'])

    @location.setter
    def location(self, v):
        if hasattr(v, 'name'):
            v = v.name
        self['location'] = v

    @property
    def next_location(self):
        branch = self.engine.branch
        turn = self.engine._things_cache.turn_after(self.character.name, self.name, *self.engine.time)
        if turn is None:
            return None
        return self.engine._get_node(self.character, self.engine._things_cache.retrieve(
            self.character.name, self.name, branch, turn, self.engine._turn_end_plan[branch, turn]
        ))

    def go_to_place(self, place, weight=''):
        """Assuming I'm in a :class:`Place` that has a :class:`Portal` direct
        to the given :class:`Place`, schedule myself to travel to the
        given :class:`Place`, taking an amount of time indicated by
        the ``weight`` stat on the :class:`Portal`, if given; else 1
        turn.

        Return the number of turns the travel will take.

        """
        if hasattr(place, 'name'):
            placen = place.name
        else:
            placen = place
        curloc = self["location"]
        orm = self.character.engine
        turns = self.engine._portal_objs[
            (self.character.name, curloc, place)].get(weight, 1)
        with self.engine.plan():
            orm.turn += turns
            self['location'] = placen
        return turns

    def follow_path(self, path, weight=None):
        """Go to several :class:`Place`s in succession, deciding how long to
        spend in each by consulting the ``weight`` stat of the
        :class:`Portal` connecting the one :class:`Place` to the next.

        Return the total number of turns the travel will take. Raise
        :class:`TravelException` if I can't follow the whole path,
        either because some of its nodes don't exist, or because I'm
        scheduled to be somewhere else.

        """
        if len(path) < 2:
            raise ValueError("Paths need at least 2 nodes")
        eng = self.character.engine
        with eng.plan():
            prevplace = path.pop(0)
            if prevplace != self['location']:
                raise ValueError("Path does not start at my present location")
            subpath = [prevplace]
            for place in path:
                if (
                        prevplace not in self.character.portal or
                        place not in self.character.portal[prevplace]
                ):
                    raise TravelException(
                        "Couldn't follow portal from {} to {}".format(
                            prevplace,
                            place
                        ),
                        path=subpath,
                        traveller=self
                    )
                subpath.append(place)
                prevplace = place
            turns_total = 0
            prevsubplace = subpath.pop(0)
            subsubpath = [prevsubplace]
            for subplace in subpath:
                portal = self.character.portal[prevsubplace][subplace]
                turn_inc = portal.get(weight, 1)
                eng.turn += turn_inc
                self.location = subplace
                turns_total += turn_inc
                subsubpath.append(subplace)
                prevsubplace = subplace
            self.location = subplace
            return turns_total

    def travel_to(self, dest, weight=None, graph=None):
        """Find the shortest path to the given :class:`Place` from where I am
        now, and follow it.

        If supplied, the ``weight`` stat of the :class:`Portal`s along
        the path will be used in pathfinding, and for deciding how
        long to stay in each Place along the way.

        The ``graph`` argument may be any NetworkX-style graph. It
        will be used for pathfinding if supplied, otherwise I'll use
        my :class:`Character`. In either case, however, I will attempt
        to actually follow the path using my :class:`Character`, which
        might not be possible if the supplied ``graph`` and my
        :class:`Character` are too different. If it's not possible,
        I'll raise a :class:`TravelException`, whose ``subpath``
        attribute holds the part of the path that I *can* follow. To
        make me follow it, pass it to my ``follow_path`` method.

        Return value is the number of turns the travel will take.

        """
        destn = dest.name if hasattr(dest, 'name') else dest
        if destn == self.location.name:
            raise ValueError("I'm already at {}".format(destn))
        graph = self.character if graph is None else graph
        path = nx.shortest_path(graph, self["location"], destn, weight)
        return self.follow_path(path, weight)


# This file is part of LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Common classes for collections in LiSE, of which most can be bound to."""
from collections import Mapping, MutableMapping
from types import MethodType
from inspect import getsource
from ast import parse, Expr, Module
import json

from blinker import Signal
from astunparse import unparse, Unparser

from .util import dedent_source


class Language(str):
    sigs = {}

    def __new__(cls, sig, v):
        me = str.__new__(cls, v)
        cls.sigs[me] = sig
        return me

    def connect(self, *args, **kwargs):
        self.sigs[self].connect(*args, **kwargs)


class AbstractLanguageDescriptor(Signal):
    def __get__(self, instance, owner=None):
        if not hasattr(self, 'lang'):
            self.lang = Language(self, self._get_language(instance))
        return self.lang

    def __set__(self, inst, val):
        self._set_language(inst, val)
        self.lang = Language(self, val)
        self.send(inst, language=val)

    def __str__(self):
        return self.lang


class LanguageDescriptor(AbstractLanguageDescriptor):
    def _get_language(self, inst):
        return inst._language

    def _set_language(self, inst, val):
        inst._language = val
        inst.query.global_set('language', val)


class StringStore(MutableMapping, Signal):
    """Store strings in database, and format them with one another upon retrieval.

    In any one string, putting the key of another string in curly
    braces will cause the other string to be substituted in.

    """
    language = LanguageDescriptor()

    def __init__(self, query, filename, lang='eng'):
        """Store the engine, the name of the database table to use, and the
        language code.

        """
        super().__init__()
        self.query = query
        self._filename = filename
        self._language = lang
        try:
            with open(filename, 'r') as inf:
                self.cache = json.load(inf)
        except FileNotFoundError:
            self.cache = {lang: {}}

    def __iter__(self):
        return iter(self.cache[self.language])

    def __len__(self):
        return len(self.cache[self.language])

    def __getitem__(self, k):
        return self.cache[self.language][k]

    def __setitem__(self, k, v):
        """Set the value of a string for the current language."""
        self.cache[self.language][k] = v
        self.send(self, key=k, val=v)

    def __delitem__(self, k):
        """Delete the string from the current language, and remove it from the
        cache.

        """
        del self.cache[self.language][k]
        self.send(self, key=k, val=None)

    def lang_items(self, lang=None):
        """Yield pairs of (id, string) for the given language."""
        if lang is None:
            lang = self.language
        yield from self.cache.setdefault(lang, {}).items()

    def save(self):
        with open(self._filename, 'w') as outf:
            json.dump(self.cache, outf, indent=4, sort_keys=True)


class FunctionStore(Signal):
    """A module-like object that lets you alter its code and save your changes.

    Instantiate it with a path to a file that you want to keep the code in.
    Assign functions to its attributes, then call its ``save()`` method,
    and they'll be unparsed and written to the file.

    This is a ``Signal``, so you can pass a function to its ``connect`` method,
    and it will be called when a function is added, changed, or deleted.
    The keyword arguments will be ``attr``, the name of the function, and ``val``,
    the function itself.

    """
    def __init__(self, filename):
        super().__init__()
        self._filename = filename
        if filename.endswith(".py"):
            filename = filename[:-3]
        try:
            with open(self._filename, 'r') as inf:
                self._ast = parse(inf.read(), filename)
                self._ast_idx = {}
                for i, node in enumerate(self._ast.body):
                    self._ast_idx[node.name] = i
                self._globl = {}
                self._locl = {}
                self._code = exec(compile(self._ast, filename, 'exec'), self._globl, self._locl)
                for thing in self._locl.values():
                    thing.__module__ = filename
        except FileNotFoundError:
            self._ast = Module(body=[])
            self._ast_idx = {}
            self._globl = {}
            self._locl = {}

    def __getattr__(self, k):
        try:
            return self._locl[k]
        except KeyError:
            raise AttributeError

    def __setattr__(self, k, v):
        if not callable(v):
            super().__setattr__(k, v)
            return
        filename = self._filename
        if filename.endswith(".py"):
            filename = filename[:-3]
        v.__module__ = filename
        self._locl[k] = v
        source = getsource(v)
        outdented = dedent_source(source)
        expr = Expr(parse(outdented))
        expr.value.body[0].name = k
        if k in self._ast_idx:
            self._ast.body[self._ast_idx[k]] = expr
        else:
            self._ast_idx[k] = len(self._ast.body)
            self._ast.body.append(expr)
        self.send(self, attr=k, val=v)

    def __call__(self, v):
        setattr(self, v.__name__, v)

    def __delattr__(self, k):
        del self._locl[k]
        del self._ast.body[self._ast_idx[k]]
        del self._ast_idx[k]
        for name in list(self._ast_idx):
            if name > k:
                self._ast_idx[name] -= 1
        self.send(self, attr=k, val=None)

    def save(self):
        with open(self._filename, 'w') as outf:
            Unparser(self._ast, outf)

    def iterplain(self):
        for funcdef in self._ast.body:
            yield funcdef.name, unparse(funcdef)

    def store_source(self, v, name=None):
        outdented = dedent_source(v)
        mod = parse(outdented)
        expr = Expr(mod)
        if len(expr.value.body) != 1:
            raise ValueError("Tried to store more than one function")
        if name is None:
            name = expr.value.body[0].name
        else:
            expr.value.body[0].name = name
        if name in self._ast_idx:
            self._ast.body[self._ast_idx[name]] = expr
        else:
            self._ast_idx[name] = len(self._ast.body)
            self._ast.body.append(expr)
        locl = {}
        exec(compile(mod, self._filename, 'exec'), {}, locl)
        self._locl.update(locl)
        self.send(self, attr=name, val=locl[name])

    def get_source(self, name):
        return unparse(self._ast.body[self._ast_idx[name]])


class MethodStore(FunctionStore):
    def __init__(self, engine):
        self.engine = engine
        super().__init__('method.py')

    def __getattr__(self, item):
        return MethodType(super().__getattr__(item), self.engine)


class UniversalMapping(MutableMapping, Signal):
    """Mapping for variables that are global but which I keep history for"""
    __slots__ = ['engine']

    def __init__(self, engine):
        """Store the engine and initialize my private dictionary of
        listeners.

        """
        super().__init__()
        self.engine = engine

    def __iter__(self):
        return self.engine._universal_cache.iter_keys(*self.engine.btt())

    def __len__(self):
        return self.engine._universal_cache.count_keys(*self.engine.btt())

    def __getitem__(self, k):
        """Get the current value of this key"""
        return self.engine._universal_cache.retrieve(k, *self.engine.btt())

    def __setitem__(self, k, v):
        """Set k=v at the current branch and tick"""
        branch, turn, tick = self.engine.nbtt()
        self.engine._universal_cache.store(k, branch, turn, tick, v)
        self.engine.query.universal_set(k, branch, turn, tick, v)
        self.engine.tick = tick
        self.send(self, key=k, val=v)

    def __delitem__(self, k):
        """Unset this key for the present (branch, tick)"""
        branch, turn, tick = self.engine.nbtt()
        self.engine._universal_cache.store(k, branch, turn, tick, None)
        self.engine.query.universal_del(k, branch, turn, tick)
        self.send(self, key=k, val=None)


class CharacterMapping(MutableMapping, Signal):
    """A mapping by which to access :class:`Character` objects.

    If a character already exists, you can always get its name here to
    get the :class:`Character` object. Deleting an item here will
    delete the character from the world, even if there are still
    :class:`Character` objects referring to it; those won't do
    anything useful anymore.

    """
    __slots__ = ['engine', '_cache']

    def __init__(self, engine):
        """Store the engine, initialize caches"""
        super().__init__()
        self.engine = engine
        self._cache = None

    def __iter__(self):
        """Iterate over every character name."""
        return iter(self.engine._graph_objs)

    def __contains__(self, name):
        """Has this character been created?"""
        if self.engine._graph_objs:
            self._cache = None
            return name in self.engine._graph_objs
        # hack to make initial load work
        if self._cache is None:
            self._cache = [ch for ch, typ in self.engine.query.graphs_types() if typ == 'DiGraph']
        return name in self._cache

    def __len__(self):
        """How many characters have been created?"""
        return len(self.engine._graph_objs)

    def __getitem__(self, name):
        """Return the named character, if it's been created.

        Try to use the cache if possible.

        """
        from .character import Character
        if name not in self:
            raise KeyError("No such character")
        cache = self.engine._graph_objs
        if name not in cache:
            cache[name] = Character(self.engine, name)
        ret = cache[name]
        if not isinstance(ret, Character):
            raise TypeError(
                """Tried to get a graph that isn't a Character.
                This should never happen. It probably indicates
                a bug in allegedb."""
            )
        return ret

    def __setitem__(self, name, value):
        """Make a new character by the given name, and initialize its data to
        the given value.

        """
        from .character import Character
        if isinstance(value, Character):
            self.engine._graph_objs[name] = value
            return
        if name in self.engine._graph_objs:
            ch = self.engine._graph_objs[name]
            ch.stat.clear()
            ch.stat.update(value)
        else:
            ch = self.engine._graph_objs[name] = Character(
                self.engine, name, data=value
            )
        self.send(self, key=name, val=ch)

    def __delitem__(self, name):
        """Delete the named character from both the cache and the database."""
        if name in self.engine._graph_objs:
            del self.engine._graph_objs[name]
        self.engine.query.del_graph(name)
        self.send(self, key=name, val=None)


class CompositeDict(Mapping):
    """Read-only mapping that looks up values in a first dict if
    available, then a second dict if possible.

    Assumes the dicts have no overlap.

    """
    __slots__ = ['d1', 'd2']

    def __init__(self, d1, d2):
        """Store dictionaries"""
        self.d1 = d1
        self.d2 = d2

    def __iter__(self):
        """Iterate over both dictionaries' keys"""
        for k in self.d1:
            yield k
        for k in self.d2:
            yield k

    def __len__(self):
        """Sum the lengths of both dictionaries"""
        return len(self.d1) + len(self.d2)

    def __getitem__(self, k):
        """Get an item from ``d1`` if possible, then ``d2``"""
        try:
            return self.d1[k]
        except KeyError:
            return self.d2[k]



# This file is part of LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" The fundamental unit of game logic, the Rule, and structures to
store and organize them in.

A Rule is three lists of functions: triggers, prereqs, and actions.
The actions do something, anything that you need your game to do, but
probably making a specific change to the world model. The triggers and
prereqs between them specify when the action should occur: any of its
triggers can tell it to happen, but then any of its prereqs may stop it
from happening.

Rules are assembled into RuleBooks, essentially just lists of Rules
that can then be assigned to be followed by any game entity --
but each game entity has its own RuleBook by default, and you never really
need to change that.

"""
from collections import (
    MutableMapping,
    MutableSequence,
    Hashable
)
from abc import ABC, abstractmethod
from functools import partial
from inspect import getsource
from ast import parse

from astunparse import unparse
from blinker import Signal

from .reify import reify
from .util import dedent_source


def roundtrip_dedent(source):
    """Reformat some lines of code into what unparse makes."""
    return unparse(parse(dedent_source(source)))


class RuleFuncList(MutableSequence, Signal):
    """Abstract class for lists of functions like trigger, prereq, action"""
    __slots__ = ['rule']

    def __init__(self, rule):
        super().__init__()
        self.rule = rule

    def _nominate(self, v):
        if callable(v):
            if hasattr(self._funcstore, v.__name__):
                stored_source = getsource(getattr(self._funcstore, v.__name__))
                new_source = getsource(v)
                if roundtrip_dedent(stored_source) != roundtrip_dedent(new_source):
                    raise KeyError(
                        "Already have a {typ} function named {n}. "
                        "If you really mean to replace it, set "
                        "engine.{typ}[{n}]".format(
                            typ=self._funcstore._filename.rstrip('.py'),
                            n=v.__name__
                        )
                    )
            else:
                self._funcstore(v)
            v = v.__name__
        if not hasattr(self._funcstore, v):
            raise KeyError("No {typ} function named {n}".format(
                typ=self._funcstore._filename.rstrip('.py'), n=v
            ))
        return v

    def _get(self):
        return self._cache.retrieve(self.rule.name, *self.rule.engine.btt())

    def _set(self, v):
        branch, turn, tick = self.rule.engine.nbtt()
        self._cache.store(self.rule.name, branch, turn, tick, v)
        self._setter(self.rule.name, branch, turn, tick, v)

    def __iter__(self):
        for funcname in self._get():
            yield getattr(self._funcstore, funcname)

    def __len__(self):
        return len(self._get())

    def __getitem__(self, i):
        return getattr(self._funcstore, self._get()[i])

    def __setitem__(self, i, v):
        v = self._nominate(v)
        l = list(self._get())
        l[i] = v
        self._set(tuple(l))
        self.send(self)

    def __delitem__(self, i):
        l = list(self._get())
        del l[i]
        self._set(tuple(l))
        self.send(self)

    def insert(self, i, v):
        l = list(self._get())
        l.insert(i, self._nominate(v))
        self._set(tuple(l))
        self.send(self)

    def append(self, v):
        self._set(self._get() + (self._nominate(v),))
        self.send(self)

    def index(self, x, start=0, end=None):
        if not callable(x):
            x = getattr(self._funcstore, x)
        return super().index(x, start, end)


class TriggerList(RuleFuncList):
    """A list of trigger functions for rules"""
    @reify
    def _funcstore(self):
        return self.rule.engine.trigger

    @reify
    def _cache(self):
        return self.rule.engine._triggers_cache

    @reify
    def _setter(self):
        return self.rule.engine.query.set_rule_triggers


class PrereqList(RuleFuncList):
    """A list of prereq functions for rules"""
    @reify
    def _funcstore(self):
        return self.rule.engine.prereq

    @reify
    def _cache(self):
        return self.rule.engine._prereqs_cache

    @reify
    def _setter(self):
        return self.rule.engine.query.set_rule_prereqs


class ActionList(RuleFuncList):
    """A list of action functions for rules"""
    @reify
    def _funcstore(self):
        return self.rule.engine.action

    @reify
    def _cache(self):
        return self.rule.engine._actions_cache

    @reify
    def _setter(self):
        return self.rule.engine.query.set_rule_actions


class RuleFuncListDescriptor(object):
    """Descriptor that lets you get and set a whole RuleFuncList at once"""
    __slots__ = ['cls']

    def __init__(self, cls):
        self.cls = cls

    @property
    def flid(self):
        return '_funclist' + str(id(self))

    def __get__(self, obj, type=None):
        if not hasattr(obj, self.flid):
            setattr(obj, self.flid, self.cls(obj))
        return getattr(obj, self.flid)

    def __set__(self, obj, value):
        if not hasattr(obj, self.flid):
            setattr(obj, self.flid, self.cls(obj))
        flist = getattr(obj, self.flid)
        namey_value = tuple(flist._nominate(v) for v in value)
        flist._set(namey_value)
        branch, turn, tick = obj.engine.nbtt()
        flist._cache.store(obj.name, branch, turn, tick, namey_value)
        flist.send(flist)

    def __delete__(self, obj):
        raise TypeError("Rules must have their function lists")


class Rule(object):
    """A collection of actions, being functions that enact some change on
    the world, which will be called each tick if and only if all of
    the prereqs return True, they being boolean functions that do not
    change the world.

    """

    triggers = RuleFuncListDescriptor(TriggerList)
    prereqs = RuleFuncListDescriptor(PrereqList)
    actions = RuleFuncListDescriptor(ActionList)

    def __init__(
            self,
            engine,
            name,
            triggers=None,
            prereqs=None,
            actions=None,
            create=True
    ):
        """Store the engine and my name, make myself a record in the database
        if needed, and instantiate one FunList each for my triggers,
        actions, and prereqs.

        """
        self.engine = engine
        self.name = self.__name__ = name
        branch, turn, tick = engine.btt()
        if create and not self.engine._triggers_cache.contains_key(name, branch, turn, tick):
            tick += 1
            self.engine.tick = tick
            triggers = tuple(self._fun_names_iter('trigger', triggers or []))
            prereqs = tuple(self._fun_names_iter('prereq', prereqs or []))
            actions = tuple(self._fun_names_iter('action', actions or []))
            self.engine.query.set_rule(
                name, branch, turn, tick, triggers, prereqs, actions
            )
            self.engine._triggers_cache.store(name, branch, turn, tick, triggers)
            self.engine._prereqs_cache.store(name, branch, turn, tick, prereqs)
            self.engine._actions_cache.store(name, branch, turn, tick, actions)

    def __eq__(self, other):
        return (
            hasattr(other, 'name') and
            self.name == other.name
        )

    def _fun_names_iter(self, functyp, val):
        """Iterate over the names of the functions in ``val``,
        adding them to ``funcstore`` if they are missing;
        or if the items in ``val`` are already the names of functions
        in ``funcstore``, iterate over those.

        """
        funcstore = getattr(self.engine, functyp)
        for v in val:
            if callable(v):
                # Overwrites anything already on the funcstore, is that bad?
                setattr(funcstore, v.__name__, v)
                yield v.__name__
            elif v not in funcstore:
                raise KeyError("Function {} not present in {}".format(
                    v, funcstore._tab
                ))
            else:
                yield v

    def __repr__(self):
        return 'Rule({})'.format(self.name)

    def trigger(self, fun):
        """Decorator to append the function to my triggers list."""
        self.triggers.append(fun)
        return fun

    def prereq(self, fun):
        """Decorator to append the function to my prereqs list."""
        self.prereqs.append(fun)
        return fun

    def action(self, fun):
        """Decorator to append the function to my actions list."""
        self.actions.append(fun)
        return fun

    def duplicate(self, newname):
        """Return a new rule that's just like this one, but under a new
        name.

        """
        if self.engine.rule.query.haverule(newname):
            raise KeyError("Already have a rule called {}".format(newname))
        return Rule(
            self.engine,
            newname,
            list(self.triggers),
            list(self.prereqs),
            list(self.actions)
        )

    def always(self):
        """Arrange to be triggered every tick, regardless of circumstance."""
        if hasattr(self.engine.trigger, 'truth'):
            truth = self.engine.trigger.truth
        else:
            def truth(*args):
                return True
        self.triggers = [truth]


class RuleBook(MutableSequence, Signal):
    """A list of rules to be followed for some Character, or a part of it
    anyway.

    """
    def _get_cache(self, branch, turn, tick):
        try:
            return self.engine._rulebooks_cache.retrieve(
                self.name, branch, turn, tick
            )
        except KeyError:
            return []

    def _set_cache(self, branch, turn, tick, v):
        self.engine._rulebooks_cache.store(self.name, branch, turn, tick, v)

    def __init__(self, engine, name):
        super().__init__()
        self.engine = engine
        self.name = name

    def __contains__(self, v):
        return getattr(v, 'name', v) in self._get_cache(*self.engine.btt())

    def __iter__(self):
        return iter(self._get_cache(*self.engine.btt()))

    def __len__(self):
        try:
            return len(self._get_cache(*self.engine.btt()))
        except KeyError:
            return 0

    def __getitem__(self, i):
        return self.engine.rule[self._get_cache(*self.engine.btt())[i]]

    def _coerce_rule(self, v):
        if isinstance(v, Rule):
            return v
        elif isinstance(v, str):
            return self.engine.rule[v]
        else:
            return Rule(self.engine, v)

    def __setitem__(self, i, v):
        v = getattr(v, 'name', v)
        branch, turn, tick = self.engine.nbtt()
        try:
            cache = self._get_cache(branch, turn, tick)
            cache[i] = v
        except KeyError:
            if i != 0:
                raise IndexError
            cache = [v]
            self._set_cache(branch, turn, tick, cache)
        self.engine.query.set_rulebook(self.name, branch, turn, tick, cache)
        self.engine._rulebooks_cache.store(self.name, branch, turn, tick, cache)
        self.engine.rulebook.send(self, i=i, v=v)
        self.send(self, i=i, v=v)

    def insert(self, i, v):
        v = getattr(v, 'name', v)
        branch, turn, tick = self.engine.nbtt()
        try:
            cache = self._get_cache(branch, turn, tick)
            cache.insert(i, v)
        except KeyError:
            if i != 0:
                raise IndexError
            cache = [v]
        self._set_cache(branch, turn, tick, cache)
        self.engine.query.set_rulebook(self.name, branch, turn, tick, cache)
        self.engine.rulebook.send(self, i=i, v=v)
        self.send(self, i=i, v=v)

    def index(self, v, start=0, stop=None):
        if isinstance(v, str):
            try:
                return self._get_cache(*self.engine.btt()).index(v, start, stop)
            except KeyError:
                raise ValueError
        return super().index(v)

    def __delitem__(self, i):
        branch, turn, tick = self.engine.btt()
        try:
            cache = self._get_cache(branch, turn, tick)
        except KeyError:
            raise IndexError
        del cache[i]
        self.engine.query.set_rulebook(self.name, branch, turn, tick, cache)
        self.engine._rulebooks_cache.store(self.name, branch, turn, tick, cache)
        self.engine.rulebook.send(self, i=i, v=None)
        self.send(self, i=i, v=None)


class RuleMapping(MutableMapping, Signal):
    """Wraps a :class:`RuleBook` so you can get its rules by name.

    You can access the rules in this either dictionary-style or as
    attributes. This is for convenience if you want to get at a rule's
    decorators, eg. to add an Action to the rule.

    Using this as a decorator will create a new rule, named for the
    decorated function, and using the decorated function as the
    initial Action.

    Using this like a dictionary will let you create new rules,
    appending them onto the underlying :class:`RuleBook`; replace one
    rule with another, where the new one will have the same index in
    the :class:`RuleBook` as the old one; and activate or deactivate
    rules. The name of a rule may be used in place of the actual rule,
    so long as the rule already exists.

    """

    def __init__(self, engine, rulebook):
        super().__init__()
        self.engine = engine
        self._rule_cache = self.engine.rule._cache
        if isinstance(rulebook, RuleBook):
            self.rulebook = rulebook
        else:
            self.rulebook = self.engine.rulebook[rulebook]

    def __repr__(self):
        return 'RuleMapping({})'.format([k for k in self])

    def __iter__(self):
        return iter(self.rulebook)

    def __len__(self):
        return len(self.rulebook)

    def __contains__(self, k):
        return k in self.rulebook

    def __getitem__(self, k):
        if k not in self:
            raise KeyError("Rule '{}' is not in effect".format(k))
        return self._rule_cache[k]

    def __getattr__(self, k):
        if k in self:
            return self[k]
        raise AttributeError

    def __setitem__(self, k, v):
        if isinstance(v, Hashable) and v in self.engine.rule:
            v = self.engine.rule[v]
        elif isinstance(v, str) and hasattr(self.engine.function, v):
            v = getattr(self.engine.function, v)
        if not isinstance(v, Rule) and callable(v):
            if k in self.engine.rule:
                raise KeyError(
                    "Already have a rule named {name}. "
                    "If you really mean to replace it, set "
                    "self.rule[{name}] to a new Rule object.".format(name=k)
                )
            # create a new rule, named k, performing action v
            self.engine.rule[k] = v
            v = self.engine.rule[k]
        assert isinstance(v, Rule)
        if isinstance(k, int):
            self.rulebook[k] = v
        else:
            self.rulebook.append(v)

    def __call__(self, v=None, name=None, always=False):
        def wrap(name, always, v):
            name = name if name is not None else v.__name__
            self[name] = v
            r = self[name]
            if always:
                r.always()
            return r
        if v is None:
            return partial(wrap, name, always)
        return wrap(name, always, v)

    def __delitem__(self, k):
        i = self.rulebook.index(k)
        del self.rulebook[i]
        self.send(self, key=k, val=None)


rule_mappings = {}
rulebooks = {}


class RuleFollower(ABC):
    """Interface for that which has a rulebook associated, which you can
    get a :class:`RuleMapping` into

    """
    __slots__ = ()

    @property
    def _rule_mapping(self):
        if id(self) not in rule_mappings:
            rule_mappings[id(self)] = self._get_rule_mapping()
        return rule_mappings[id(self)]

    # keeping _rulebooks out of the instance lets subclasses
    # use __slots__ without having _rulebooks in the slots
    @property
    def _rulebooks(self):
        return rulebooks[id(self)]

    @_rulebooks.setter
    def _rulebooks(self, v):
        rulebooks[id(self)] = v

    @property
    def rule(self, v=None, name=None):
        if v is not None:
            return self._rule_mapping(v, name)
        return self._rule_mapping

    @property
    def rulebook(self):
        if not hasattr(self, '_rulebook'):
            self._upd_rulebook()
        return self._rulebook

    @rulebook.setter
    def rulebook(self, v):
        n = v.name if isinstance(v, RuleBook) else v
        try:
            if n == self._get_rulebook_name():
                return
        except KeyError:
            pass
        self._set_rulebook_name(n)
        self._upd_rulebook()

    def _upd_rulebook(self):
        self._rulebook = self._get_rulebook()

    def _get_rulebook(self):
        return self.engine.rulebook[self._get_rulebook_name()]

    def rules(self):
        if not hasattr(self, 'engine'):
            raise AttributeError("Need an engine before I can get rules")
        return self._rule_mapping.values()

    @abstractmethod
    def _get_rule_mapping(self):
        """Get the :class:`RuleMapping` for my rulebook."""
        raise NotImplementedError

    @abstractmethod
    def _get_rulebook_name(self):
        """Get the name of my rulebook."""
        raise NotImplementedError

    @abstractmethod
    def _set_rulebook_name(self, n):
        """Tell the database that this is the name of the rulebook to use for
        me.

        """
        raise NotImplementedError


class AllRuleBooks(MutableMapping, Signal):
    __slots__ = ['engine', '_cache']

    def __init__(self, engine):
        super().__init__()
        self.engine = engine
        self._cache = {}

    def __iter__(self):
        return self.engine._rulebooks_cache.iter_entities(*self.engine.btt())

    def __len__(self):
        return len(list(self))

    def __contains__(self, k):
        return self.engine._rulebooks_cache.contains_entity(k, *self.engine.btt())

    def __getitem__(self, k):
        if k not in self._cache:
            self._cache[k] = RuleBook(self.engine, k)
        return self._cache[k]

    def __setitem__(self, key, value):
        if key not in self._cache:
            self._cache[key] = RuleBook(self.engine, key)
        rb = self._cache[key]
        while len(rb) > 0:
            del rb[0]
        rb.extend(value)

    def __delitem__(self, key):
        self.engine._del_rulebook(key)


class AllRules(MutableMapping, Signal):
    """A mapping of every rule in the game.

    You can use this as a decorator to make a rule and not assign it
    to anything.

    """
    def __init__(self, engine):
        super().__init__()
        self.engine = engine

    @reify
    def _cache(self):
        return self.engine._rules_cache

    def __iter__(self):
        yield from self._cache

    def __len__(self):
        return len(self._cache)

    def __contains__(self, k):
        return k in self._cache

    def __getitem__(self, k):
        return self._cache[k]

    def __setitem__(self, k, v):
        # you can use the name of a stored function or rule
        if isinstance(v, str):
            if hasattr(self.engine.action, v):
                v = getattr(self.engine.action, v)
            elif hasattr(self.engine.function, v):
                v = getattr(self.engine.function, v)
            elif hasattr(self.engine.rule, v):
                v = getattr(self.engine.rule, v)
            else:
                raise ValueError("Unknown function: " + v)
        if callable(v):
            if k not in self._cache:
                self._cache[k] = Rule(self.engine, k, actions=[v])
                new = self._cache[k]
            else:
                new = self._cache[k]
                new.actions = [v]
        elif isinstance(v, Rule):
            self._cache[k] = v
            new = v
        else:
            raise TypeError(
                "Don't know how to store {} as a rule.".format(type(v))
            )
        self.send(self, key=new, rule=v)

    def __delitem__(self, k):
        if k not in self:
            raise KeyError("No such rule")
        for rulebook in self.engine.rulebooks.values():
            try:
                del rulebook[rulebook.index(k)]
            except IndexError:
                pass
        del self._cache[k]
        self.send(self, key=k, rule=None)

    def __call__(self, v=None, name=None):
        if v is None and name is not None:
            def r(f):
                self[name] = f
                return self[name]
            return r
        k = name if name is not None else v.__name__
        self[k] = v
        return self[k]

    def new_empty(self, name):
        """Make a new rule with no actions or anything, and return it."""
        if name in self:
            raise KeyError("Already have rule {}".format(name))
        new = Rule(self.engine, name)
        self._cache[name] = new
        self.send(self, rule=new, active=True)
        return new



# This file is part of LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import cherrypy
import threading
import logging
from queue import Queue
from ..handle import EngineHandle


class LiSEHandleWebService(object):
    exposed = True

    def __init__(self, *args, **kwargs):
        if 'logger' in kwargs:
            self.logger = kwargs['logger']
        else:
            self.logger = kwargs['logger'] = logging.getLogger(__name__)
        self.cmdq = kwargs['cmdq'] = Queue()
        self.outq = kwargs['outq'] = Queue()
        self._handle_thread = threading.Thread(
            target=self._run_handle_forever, args=args, kwargs=kwargs,
            daemon=True
        )
        self._handle_thread.start()

    @staticmethod
    def _run_handle_forever(*args, **kwargs):
        cmdq = kwargs.pop('cmdq')
        outq = kwargs.pop('outq')
        logger = kwargs.pop('logger')
        setup = kwargs.pop('setup', None)
        logq = Queue()

        def log(typ, data):
            if typ == 'command':
                (cmd, args) = data
                logger.debug(
                    "LiSE thread {}: calling {}{}".format(
                        threading.get_ident(),
                        cmd,
                        tuple(args)
                    )
                )
            else:
                logger.debug(
                    "LiSE thread {}: returning {} (of type {})".format(
                        threading.get_ident(),
                        data,
                        repr(type(data))
                    )
                )

        def get_log_forever(logq):
            (level, data) = logq.get()
            getattr(logger, level)(data)

        engine_handle = EngineHandle(args, kwargs, logq)
        if setup:
            setup(engine_handle._real)
        handle_log_thread = threading.Thread(
            target=get_log_forever, args=(logq,), daemon=True
        )
        handle_log_thread.start()
        while True:
            inst = cmdq.get()
            if inst == 'shutdown':
                handle_log_thread.join()
                cmdq.close()
                outq.close()
                return 0
            cmd = inst.pop('command')
            silent = inst.pop('silent', False)
            log('command', (cmd, args))
            response = getattr(engine_handle, cmd)(**inst)
            if silent:
                continue
            log('result', response)
            outq.put(engine_handle._real.listify(response))

    @cherrypy.tools.accept(media='application/json')
    @cherrypy.tools.json_out()
    def GET(self):
        return cherrypy.session['LiSE_response']

    @cherrypy.tools.json_out()
    def POST(self, **kwargs):
        silent = kwargs.get('silent', False)
        self.cmdq.put(kwargs)
        if silent:
            return None
        response = self.outq.get()
        cherrypy.session['LiSE_response'] = response
        return response

    def PUT(self, silent=False, **kwargs):
        silent = silent
        self.cmdq.put(kwargs)
        if not silent:
            cherrypy.session['LiSE_response'] = self.outq.get()

    def DELETE(self):
        cherrypy.session.pop('LiSE_response', None)



# This file is part of LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from allegedb.cache import (
    Cache,
    PickyDefaultDict,
    StructuredDefaultDict,
    TurnDict,
    HistoryError
)
from .util import singleton_get, sort_set


class InitializedCache(Cache):
    def _store_journal(self, *args):
        entity, key, branch, turn, tick, value = args[-6:]
        parent = args[:-6]
        settings_turns = self.settings[branch]
        presettings_turns = self.presettings[branch]
        try:
            prev = self.retrieve(*args[:-1])
        except KeyError:
            return  # because you can't rewind past this
        if prev == value:
            return  # not much point reporting on a non-change in a diff
        if turn in settings_turns or turn in settings_turns.future():
            assert turn in presettings_turns or turn in presettings_turns.future()
            setticks = settings_turns[turn]
            presetticks = presettings_turns[turn]
            presetticks[tick] = parent + (entity, key, prev)
            setticks[tick] = parent + (entity, key, value)
        else:
            presettings_turns[turn] = {tick: parent + (entity, key, prev)}
            settings_turns[turn] = {tick: parent + (entity, key, value)}


class EntitylessCache(Cache):
    def store(self, key, branch, turn, tick, value, *, planning=None):
        super().store(None, key, branch, turn, tick, value, planning=planning)

    def load(self, data, validate=False):
        return super().load(((None,) + row for row in data), validate)

    def retrieve(self, key, branch, turn, tick):
        return super().retrieve(None, key, branch, turn, tick)

    def iter_entities_or_keys(self, branch, turn, tick, *, forward=None):
        return super().iter_entities_or_keys(None, branch, turn, tick, forward=forward)
    iter_entities = iter_keys = iter_entities_or_keys

    def contains_entity_or_key(self, ke, branch, turn, tick):
        return super().contains_entity_or_key(None, ke, branch, turn, tick)
    contains_entity = contains_key = contains_entity_or_key

    def retrieve(self, *args):
        return super().retrieve(*(None,)+args)


class InitializedEntitylessCache(EntitylessCache, InitializedCache):
    pass


class AvatarnessCache(Cache):
    """A cache for remembering when a node is an avatar of a character."""
    def __init__(self, engine):
        Cache.__init__(self, engine)
        self.user_order = StructuredDefaultDict(3, TurnDict)
        self.user_shallow = PickyDefaultDict(TurnDict)
        self.graphs = StructuredDefaultDict(1, TurnDict)
        self.graphavs = StructuredDefaultDict(1, TurnDict)
        self.charavs = StructuredDefaultDict(1, TurnDict)
        self.soloav = StructuredDefaultDict(1, TurnDict)
        self.uniqav = StructuredDefaultDict(1, TurnDict)
        self.uniqgraph = StructuredDefaultDict(1, TurnDict)
        self.users = StructuredDefaultDict(1, TurnDict)

    def _store(self, character, graph, node, branch, turn, tick, is_avatar, *, planning):
        is_avatar = True if is_avatar else None
        super()._store(character, graph, node, branch, turn, tick, is_avatar, planning=planning)
        userturns = self.user_order[graph][node][character][branch]
        if turn in userturns:
            userturns[turn][tick] = is_avatar
        else:
            userturns[turn] = {tick: is_avatar}
        usershal = self.user_shallow[(graph, node, character, branch)]
        if turn in usershal:
            usershal[turn][tick] = is_avatar
        else:
            usershal[turn] = {tick: is_avatar}
        charavs = self.charavs[character][branch]
        graphavs = self.graphavs[(character, graph)][branch]
        graphs = self.graphs[character][branch]
        uniqgraph = self.uniqgraph[character][branch]
        soloav = self.soloav[(character, graph)][branch]
        uniqav = self.uniqav[character][branch]
        users = self.users[graph, node][branch]

        def add_something(cache, what):
            if turn in cache:
                nucache = cache[turn][tick].copy()
                nucache.add(what)
                cache[turn][tick] = nucache
            elif cache.rev_gettable(turn):
                cacheturn = cache[turn]
                nucache = cacheturn[cacheturn.end].copy()
                nucache.add(what)
                cache[turn] = {tick: nucache}
            else:
                cache[turn] = {tick: {what}}

        def remove_something(cache, what):
            if turn in cache:
                nucache = cache[turn][tick].copy()
                nucache.remove(what)
                cache[turn][tick] = nucache
            elif cache.rev_gettable(turn):
                cacheturn = cache[turn]
                nucache = cacheturn[cacheturn.end].copy()
                nucache.remove(what)
                cache[turn] = {tick: nucache}
            else:
                raise ValueError
        if is_avatar:
            add_something(graphavs, node)
            add_something(charavs, (graph, node))
            add_something(graphs, graph)
            add_something(users, character)
        else:
            remove_something(graphavs, node)
            remove_something(charavs, (graph, node))
            if not graphavs[turn][tick]:
                remove_something(graphs, graph)
            if not charavs[turn][tick]:
                remove_something(users, character)
        graphav = singleton_get(graphavs[turn][tick])
        if turn in soloav:
            soloav[turn][tick] = graphav
        else:
            soloav[turn] = {tick: graphav}
        charav = singleton_get(charavs[turn][tick])
        if turn in uniqav:
            uniqav[turn][tick] = charav
        else:
            uniqav[turn] = {tick: charav}
        if not graphavs[turn][tick]:
            graphs[turn][tick].remove(graph)
            if len(graphs[turn][tick]) == 1:
                uniqgraph[turn][tick] = next(iter(graphs[turn][tick]))
            else:
                uniqgraph[turn][tick] = None
        if turn in graphavs and tick in graphavs[turn] and len(graphavs[turn][tick]) != 1:
            if turn in soloav:
                soloav[turn][tick] = None
            else:
                soloav[turn] = soloav.cls({tick: None})
        else:
            if turn in soloav:
                soloav[turn][tick] = node
            else:
                soloav[turn] = soloav.cls({tick: None})
        if turn in charavs and charavs[turn].rev_gettable(tick) and len(charavs[turn][tick]) != 1:
            if turn in uniqav:
                uniqav[turn][tick] = None
            else:
                uniqav[turn] = uniqav.cls({tick: None})
        elif turn in uniqav:
            uniqav[turn][tick] = (graph, node)
        else:
            uniqav[turn] = uniqav.cls({tick: (graph, node)})
        if turn in graphs and graphs[turn].rev_gettable(tick) and len(graphs[turn][tick]) != 1:
            if turn in uniqgraph:
                uniqgraph[turn][tick] = None
            else:
                uniqgraph[turn] = uniqgraph.cls({tick: None})
        elif turn in uniqgraph:
            uniqgraph[turn][tick] = graph
        else:
            uniqgraph[turn] = uniqgraph.cls({tick: graph})

    def get_char_graph_avs(self, char, graph, branch, turn, tick):
        return self._valcache_lookup(
            self.graphavs[(char, graph)], branch, turn, tick
        ) or set()

    def get_char_graph_solo_av(self, char, graph, branch, turn, tick):
        return self._valcache_lookup(
            self.soloav[(char, graph)], branch, turn, tick
        )

    def get_char_only_av(self, char, branch, turn, tick):
        return self._valcache_lookup(
            self.uniqav[char], branch, turn, tick
        )

    def get_char_only_graph(self, char, branch, turn, tick):
        return self._valcache_lookup(
            self.uniqgraph[char], branch, turn, tick
        )

    def get_char_graphs(self, char, branch, turn, tick):
        return self._valcache_lookup(
            self.graphs[char], branch, turn, tick
        ) or set()

    def _slow_iter_character_avatars(self, character, branch, turn, tick, *, forward):
        for graph in self.iter_entities(character, branch, turn, tick, forward=forward):
            for node in self.iter_entities(character, graph, branch, turn, tick, forward=forward):
                yield graph, node

    def _slow_iter_users(self, graph, node, branch, turn, tick):
        if graph not in self.user_order:
            return
        for character in self.user_order[graph][node]:
            if (graph, node, character, branch) not in self.user_shallow:
                for (b, t, tc) in self.db._iter_parent_btt(branch, turn, tick):
                    if b in self.user_order[graph][node][character]:
                        isav = self.user_order[graph][node][character][b][t]
                        # side effect!! bad!
                        self.store(character, graph, node, branch, turn, tick, isav[tc])
                        break
                else:
                    self.store(character, graph, node, branch, turn, tick, None)
            try:
                if self.user_shallow[(graph, node, character, branch)][turn][tick]:
                   yield character
            except HistoryError:
                continue


class RulesHandledCache(object):
    def __init__(self, engine):
        self.engine = engine
        self.handled = {}
        self.unhandled = {}

    def get_rulebook(self, *args):
        raise NotImplementedError

    def iter_unhandled_rules(self, branch, turn, tick):
        raise NotImplementedError

    def store(self, *args, loading=False):
        entity = args[:-5]
        rulebook, rule, branch, turn, tick = args[-5:]
        self.handled.setdefault(entity + (rulebook, branch, turn), set()).add(rule)
        unhandl = self.unhandled.setdefault(entity, {}).setdefault(rulebook, {}).setdefault(branch, {})
        if turn not in unhandl:
            unhandl[turn] = list(self.unhandled_rulebook_rules(entity, rulebook, branch, turn, tick))
        try:
            unhandl[turn].remove(rule)
        except ValueError:
            if not loading:
                raise

    def fork(self, branch, turn, tick):
        parent_branch, parent_turn, parent_tick, end_turn, end_tick = self.engine._branches[branch]
        unhandl = self.unhandled
        handl = self.handled
        rbcache = self.engine._rulebooks_cache
        unhandrr = self.unhandled_rulebook_rules
        getrb = self.get_rulebook
        for entity in unhandl:
            rulebook = getrb(*entity + (branch, turn, tick))
            if entity + (rulebook, branch, turn) in handl:
                raise HistoryError(
                    "Tried to fork history in a RulesHandledCache, "
                    "but it seems like rules have already been run where we're "
                    "forking to"
                )
            rules = set(rbcache.retrieve(rulebook, branch, turn, tick))
            unhandled_rules = unhandrr(entity, rulebook, parent_branch, parent_turn, tick)
            unhandled_rules_set = set(unhandled_rules)
            handl[entity + (rulebook, branch, turn)] = rules.difference(unhandled_rules_set)
            unhandl[entity][rulebook][branch][turn] = unhandled_rules

    def retrieve(self, *args):
        return self.handled[args]

    def unhandled_rulebook_rules(self, *args):
        entity = args[:-4]
        rulebook, branch, turn, tick = args[-4:]
        if (
            entity in self.unhandled and
            rulebook in self.unhandled[entity] and
            branch in self.unhandled[entity][rulebook] and
            turn in self.unhandled[entity][rulebook][branch]
        ):
            ret = self.unhandled[entity][rulebook][branch][turn]
        else:
            try:
                return [
                    rule for rule in
                    self.engine._rulebooks_cache.retrieve(rulebook, branch, turn, tick)
                    if rule not in self.handled.setdefault(entity + (rulebook, branch, turn), set())
                ]
            except KeyError:
                return []
        return ret


class CharacterRulesHandledCache(RulesHandledCache):
    def get_rulebook(self, character, branch, turn, tick):
        try:
            return self.engine._characters_rulebooks_cache.retrieve(character, branch, turn, tick)
        except KeyError:
            return character, 'character'

    def iter_unhandled_rules(self, branch, turn, tick):
        for character in sort_set(self.engine.character.keys()):
            rb = self.get_rulebook(character, branch, turn, tick)
            for rule in self.unhandled_rulebook_rules(
                character, rb, branch, turn, tick
            ):
                yield character, rb, rule


class AvatarRulesHandledCache(RulesHandledCache):
    def get_rulebook(self, character, branch, turn, tick):
        try:
            return self.engine._avatars_rulebooks_cache.retrieve(character, branch, turn, tick)
        except KeyError:
            return character, 'avatar'

    def iter_unhandled_rules(self, branch, turn, tick):
        charm = self.engine.character
        for character in sort_set(charm.keys()):
            rulebook = self.get_rulebook(character, branch, turn, tick)
            charavm = charm[character].avatar
            for graph in sort_set(charavm.keys()):
                for avatar in sort_set(charavm[graph].keys()):
                    try:
                        rules = self.unhandled_rulebook_rules(character, graph, avatar, rulebook, branch, turn, tick)
                    except KeyError:
                        continue
                    for rule in rules:
                        yield character, graph, avatar, rulebook, rule


class CharacterThingRulesHandledCache(RulesHandledCache):
    def get_rulebook(self, character, branch, turn, tick):
        try:
            return self.engine._characters_things_rulebooks_cache.retrieve(character, branch, turn, tick)
        except KeyError:
            return character, 'character_thing'

    def iter_unhandled_rules(self, branch, turn, tick):
        charm = self.engine.character
        for character in sort_set(charm.keys()):
            rulebook = self.get_rulebook(character, branch, turn, tick)
            things = sort_set(charm[character].thing.keys())
            pass
            for thing in things:
                try:
                    rules = self.unhandled_rulebook_rules(character, thing, rulebook, branch, turn, tick)
                except KeyError:
                    continue
                for rule in rules:
                    yield character, thing, rulebook, rule


class CharacterPlaceRulesHandledCache(RulesHandledCache):
    def get_rulebook(self, character, branch, turn, tick):
        try:
            return self.engine._characters_places_rulebooks_cache.retrieve(character, branch, turn, tick)
        except KeyError:
            return character, 'character_place'

    def iter_unhandled_rules(self, branch, turn, tick):
        charm = self.engine.character
        for character in sort_set(charm.keys()):
            rulebook = self.get_rulebook(character, branch, turn, tick)
            for place in sort_set(charm[character].place.keys()):
                try:
                    rules = self.unhandled_rulebook_rules(character, place, rulebook, branch, turn, tick)
                except KeyError:
                    continue
                for rule in rules:
                    yield character, place, rulebook, rule


class CharacterPortalRulesHandledCache(RulesHandledCache):
    def get_rulebook(self, character, branch, turn, tick):
        try:
            return self.engine._characters_portals_rulebooks_cache.retrieve(character, branch, turn, tick)
        except KeyError:
            return character, 'character_portal'

    def iter_unhandled_rules(self, branch, turn, tick):
        charm = self.engine.character
        for character in sort_set(charm.keys()):
            try:
                rulebook = self.get_rulebook(character, branch, turn, tick)
            except KeyError:
                continue
            charp = charm[character].portal
            for orig in sort_set(charp.keys()):
                for dest in sort_set(charp[orig].keys()):
                    try:
                        rules = self.unhandled_rulebook_rules(character, orig, dest, rulebook, branch, turn, tick)
                    except KeyError:
                        continue
                    for rule in rules:
                        yield character, orig, dest, rulebook, rule


class NodeRulesHandledCache(RulesHandledCache):
    def get_rulebook(self, character, node, branch, turn, tick):
        try:
            return self.engine._nodes_rulebooks_cache.retrieve(character, node, branch, turn, tick)
        except KeyError:
            return character, node

    def iter_unhandled_rules(self, branch, turn, tick):
        charm = self.engine.character
        for character in sort_set(charm.keys()):
            for node in sort_set(charm[character].node.keys()):
                try:
                    rulebook = self.get_rulebook(character, node, branch, turn, tick)
                    rules = self.unhandled_rulebook_rules(character, node, rulebook, branch, turn, tick)
                except KeyError:
                    continue
                for rule in rules:
                    yield character, node, rulebook, rule


class PortalRulesHandledCache(RulesHandledCache):
    def get_rulebook(self, character, orig, dest, branch, turn, tick):
        try:
            return self.engine._portals_rulebooks_cache.retrieve(character, orig, dest, branch, turn, tick)
        except KeyError:
            return character, orig, dest

    def iter_unhandled_rules(self, branch, turn, tick):
        charm = self.engine.character
        for character in sort_set(charm.keys()):
            charp = charm[character].portal
            for orig in sort_set(charp.keys()):
                dests = charp[orig]
                for dest in sort_set(dests.keys()):
                    try:
                        rulebook = self.get_rulebook(character, orig, dest, branch, turn, tick)
                        rules = self.unhandled_rulebook_rules(character, orig, dest, rulebook, branch, turn, tick)
                    except KeyError:
                        continue
                    for rule in rules:
                        yield character, orig, dest, rulebook, rule


class ThingsCache(Cache):
    def __init__(self, db):
        Cache.__init__(self, db)
        self._make_node = db.thing_cls

    def _store(self, *args, planning):
        character, thing, branch, turn, tick, location = args
        try:
            oldloc = self.retrieve(character, thing, branch, turn, tick)
        except KeyError:
            oldloc = None
        super()._store(*args, planning=planning)
        if oldloc is not None:
            oldnodecont = self.db._node_contents_cache.retrieve(
                character, oldloc, branch, turn, tick
            )
            self.db._node_contents_cache.store(
                character, thing, branch, turn, tick, oldnodecont.difference((thing,))
            )
        try:
            newnodecont = self.db._node_contents_cache.retrieve(
                character, location, branch, turn, tick
            )
        except KeyError:
            newnodecont = frozenset()
        self.db._node_contents_cache.store(
            character, location, branch, turn, tick,
            newnodecont.union((thing,))
        )

    def turn_before(self, character, thing, branch, turn):
        try:
            self.retrieve(character, thing, branch, turn, 0)
        except KeyError:
            pass
        return self.keys[(character,)][thing][branch].rev_before(turn)

    def turn_after(self, character, thing, branch, turn):
        try:
            self.retrieve(character, thing, branch, turn, 0)
        except KeyError:
            pass
        return self.keys[(character,)][thing][branch].rev_after(turn)


# This file is part of LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""The top level of the LiSE world model, the Character.

Based on NetworkX DiGraph objects with various additions and
conveniences.

A Character is a graph that follows rules. Its rules may be assigned
to run on only some portion of it: just edges (called Portals), just
nodes, or just nodes of the kind that have a location in another node
(called Places and Things, respectively). Each Character has a
``stat`` property that acts very much like a dictionary, in which you
can store game-relevant data for the rules to use.

You can designate some nodes in one Character as avatars of another,
and then assign a rule to run on all of a Character's avatars. This is
useful for the common case where someone in your game has a location
in the physical world (here, a Character, called 'physical') but also
has a behavior flowchart, or a skill tree, that isn't part of the
physical world. In that case the flowchart is the person's Character,
and their node in the physical world is an avatar of it.

"""

from abc import abstractmethod
from collections import (
    Mapping,
    MutableMapping,
    Callable
)
from operator import ge, gt, le, lt, eq
from blinker import Signal

import networkx as nx
from allegedb.graph import (
    DiGraph,
    GraphNodeMapping,
    GraphSuccessorsMapping,
    DiGraphPredecessorsMapping
)
from allegedb.wrap import MutableMappingUnwrapper

from .xcollections import CompositeDict
from .rule import RuleMapping
from .rule import RuleFollower as BaseRuleFollower
from .node import Node
from .thing import Thing
from .place import Place
from .portal import Portal
from .util import getatt, singleton_get
from .exc import AmbiguousAvatarError, WorldIntegrityError


class SpecialMappingDescriptor:
    def __init__(self, mapclsname):
        self.mapps = {}
        self.mapclsname = mapclsname

    def __get__(self, instance, owner):
        if id(instance) in self.mapps:
            return self.mapps[id(instance)]
        mappcls = getattr(instance, self.mapclsname)
        ret = self.mapps[id(instance)] = mappcls(instance)
        return ret

    def __set__(self, instance, value):
        if id(instance) not in self.mapps:
            self.mapps[id(instance)] = getattr(instance, self.mapclsname)(instance)
        it = self.mapps[id(instance)]
        it.clear()
        it.update(value)


class AbstractCharacter(MutableMapping):

    """The Character API, with all requisite mappings and graph generators.

    Mappings resemble those of a NetworkX digraph:

    * ``thing`` and ``place`` are subsets of ``node``
    * ``edge``, ``adj``, and ``succ`` are aliases of ``portal``
    * ``pred`` is an alias to ``preportal``
    * ``stat`` is a dict-like mapping of data that changes over game-time,
    to be used in place of graph attributes

    """
    engine = getatt('db')
    no_unwrap = True

    @abstractmethod
    def add_place(self, name, **kwargs): pass

    def add_node(self, name, **kwargs):
        self.add_place(name, **kwargs)

    @abstractmethod
    def add_places_from(self, seq, **attrs): pass

    def add_nodes_from(self, seq, **attrs):
        self.add_places_from(seq, **attrs)

    def new_place(self, name, **kwargs):
        if name not in self.node:
            self.add_place(name, **kwargs)
            return self.place[name]
        if isinstance(name, str):
            n = 0
            while name + str(n) in self.node:
                n += 1
            self.add_place(name + str(n), **kwargs)
            return self.place[name]
        raise KeyError("Already have a node named {}".format(name))

    def new_node(self, name, **kwargs):
        return self.new_place(name, **kwargs)

    @abstractmethod
    def add_thing(self, name, location, **kwargs): pass

    @abstractmethod
    def add_things_from(self, seq, **attrs): pass

    def new_thing(
            self, name, location, **kwargs
    ):
        if name not in self.node:
            self.add_thing(name, location, **kwargs)
            return self.thing[name]
        if isinstance(name, str):
            n = 0
            while name + str(n) in self.node:
                n += 1
            self.add_thing(name + str(n), location, **kwargs)
            return self.thing[name]
        raise KeyError("Already have a thing named {}".format(name))

    @abstractmethod
    def thing2place(self, name): pass

    @abstractmethod
    def place2thing(self, name, location): pass

    @abstractmethod
    def add_portal(self, orig, dest, symmetrical=False, **kwargs): pass

    def add_edge(self, orig, dest, **kwargs):
        self.add_portal(orig, dest, **kwargs)

    def new_portal(self, orig, dest, symmetrical=False, **kwargs):
        self.add_portal(orig, dest, symmetrical, **kwargs)
        return self.portal[orig][dest]

    @abstractmethod
    def add_portals_from(self, seq, **attrs): pass

    def add_edges_from(self, seq, **attrs):
        self.add_portals_from(seq, **attrs)

    def remove_portals_from(self, seq):
        for orig, dest in seq:
            del self.portal[orig][dest]

    def remove_edges_from(self, seq):
        self.remove_portals_from(seq)

    @abstractmethod
    def add_avatar(self, a, b=None): pass

    @abstractmethod
    def del_avatar(self, a, b=None): pass

    def __eq__(self, other):
        return isinstance(other, AbstractCharacter) and self.name == other.name

    def __iter__(self):
        return iter(self.node)

    def __len__(self):
        return len(self.node)

    def __contains__(self, k):
        return k in self.node

    def __getitem__(self, k):
        return self.node[k]

    def __setitem__(self, k, v):
        self.node[k] = v

    def __delitem__(self, k):
        del self.node[k]

    thing = SpecialMappingDescriptor('ThingMapping')
    place = SpecialMappingDescriptor('PlaceMapping')
    node = _node = SpecialMappingDescriptor('ThingPlaceMapping')
    portal = adj = succ = edge = _adj = _succ = SpecialMappingDescriptor('PortalSuccessorsMapping')
    preportal = pred = _pred = SpecialMappingDescriptor('PortalPredecessorsMapping')
    avatar = SpecialMappingDescriptor('AvatarGraphMapping')
    stat = getatt('graph')

    def historical(self, stat):
        from .query import StatusAlias
        return StatusAlias(
            entity=self.stat,
            stat=stat
        )

    def do(self, func, *args, **kwargs):
        """Apply the function to myself, and return myself.

        Look up the function in the database if needed. Pass it any
        arguments given, keyword or positional.

        Useful chiefly when chaining.

        """
        if not callable(func):
            func = getattr(self.engine.function, func)
        func(self, *args, **kwargs)
        return self

    def perlin(self, stat='perlin'):
        """Apply Perlin noise to my nodes, and return myself.

        I'll try to use the name of the node as its spatial position
        for this purpose, or use its stats 'x', 'y', and 'z', or skip
        the node if neither are available. z is assumed 0 if not
        provided for a node.

        Result will be stored in a node stat named 'perlin' by default.
        Supply the name of another stat to use it instead.

        """
        from math import floor
        p = self.engine.shuffle([
            151, 160, 137, 91, 90, 15, 131, 13, 201, 95, 96, 53, 194, 233, 7,
            225, 140, 36, 103, 30, 69, 142, 8, 99, 37, 240, 21, 10, 23, 190,
            6, 148, 247, 120, 234, 75, 0, 26, 197, 62, 94, 252, 219, 203, 117,
            35, 11, 32, 57, 177, 33, 88, 237, 149, 56, 87, 174, 20, 125, 136,
            171, 168, 68, 175, 74, 165, 71, 134, 139, 48, 27, 166, 77, 146,
            158, 231, 83, 111, 229, 122, 60, 211, 133, 230, 220, 105, 92, 41,
            55, 46, 245, 40, 244, 102, 143, 54, 65, 25, 63, 161, 1, 216, 80,
            73, 209, 76, 132, 187, 208, 89, 18, 169, 200, 196, 135, 130, 116,
            188, 159, 86, 164, 100, 109, 198, 173, 186, 3, 64, 52, 217, 226,
            250, 124, 123, 5, 202, 38, 147, 118, 126, 255, 82, 85, 212, 207,
            206, 59, 227, 47, 16, 58, 17, 182, 189, 28, 42, 223, 183, 170, 213,
            119, 248, 152, 2, 44, 154, 163, 70, 221, 153, 101, 155, 167, 43,
            172, 9, 129, 22, 39, 253, 19, 98, 108, 110, 79, 113, 224, 232, 178,
            185, 112, 104, 218, 246, 97, 228, 251, 34, 242, 193, 238, 210, 144,
            12, 191, 179, 162, 241, 81, 51, 145, 235, 249, 14, 239, 107, 49,
            192, 214, 31, 181, 199, 106, 157, 184, 84, 204, 176, 115, 121, 50,
            45, 127, 4, 150, 254, 138, 236, 205, 93, 222, 114, 67, 29, 24, 72,
            243, 141, 128, 195, 78, 66, 215, 61, 156, 180
        ]) * 2

        def fade(t):
            return t * t * t * (t * (t * 6 - 15) + 10)

        def lerp(t, a, b):
            return a + t * (b - a)

        def grad(hsh, x, y, z):
            """CONVERT LO 4 BITS OF HASH CODE INTO 12 GRADIENT DIRECTIONS."""
            h = hsh & 15
            u = x if h < 8 else y
            v = y if h < 4 else x if h == 12 or h == 14 else z
            return (u if h & 1 == 0 else -u) + (v if h & 2 == 0 else -v)

        def noise(x, y, z):
            # FIND UNIT CUBE THAT CONTAINS POINT.
            X = int(x) & 255
            Y = int(y) & 255
            Z = int(z) & 255
            # FIND RELATIVE X, Y, Z OF POINT IN CUBE.
            x -= floor(x)
            y -= floor(y)
            z -= floor(z)
            # COMPUTE FADE CURVES FOR EACH OF X, Y, Z.
            u = fade(x)
            v = fade(y)
            w = fade(z)
            # HASH COORDINATES OF THE 8 CUBE CORNERS,
            A = p[X] + Y
            AA = p[A] + Z
            AB = p[A+1] + Z
            B = p[X+1] + y
            BA = p[B] + Z
            BB = p[B+1] + Z
            # AND ADD BLENDED RESULTS FROM 8 CORNERS OF CUBE
            return lerp(
                w,
                lerp(
                    v,
                    lerp(
                        u,
                        grad(p[AA], x, y, z),
                        grad(p[BA], x-1, y, z)
                    ),
                    lerp(
                        u,
                        grad(p[AB], x, y-1, z),
                        grad(p[BB], x-1, y-1, z)
                    )
                ),
                lerp(
                    v,
                    lerp(
                        u,
                        grad(p[AA+1], x, y, z-1),
                        grad(p[BA+1], x-1, y, z-1)
                    ),
                    lerp(
                        u,
                        grad(p[AB+1], x, y-1, z-1),
                        grad(p[BB+1], x-1, y-1, z-1)
                    )
                )
            )

        for node in self.node.values():
            try:
                (x, y, z) = node.name
            except ValueError:
                try:
                    (x, y) = node.name
                    z = 0.0
                except ValueError:
                    try:
                        x = node['x']
                        y = node['y']
                        z = node.get('z', 0.0)
                    except KeyError:
                        continue
            x, y, z = map(float, (x, y, z))
            node[stat] = noise(x, y, z)

        return self

    def copy_from(self, g):
        """Copy all nodes and edges from the given graph into this.

        Return myself.

        """
        renamed = {}
        for k, v in g.node.items():
            ok = k
            if k in self.place:
                n = 0
                while k in self.place:
                    k = ok + (n,) if isinstance(ok, tuple) else (ok, n)
                    n += 1
            renamed[ok] = k
            self.place[k] = v
        if type(g) is nx.MultiDiGraph:
            g = nx.DiGraph(g)
        elif type(g) is nx.MultiGraph:
            g = nx.Graph(g)
        if type(g) is nx.DiGraph:
            for u, v in g.edges:
                self.edge[renamed[u]][renamed[v]] = g.adj[u][v]
        else:
            assert type(g) is nx.Graph
            for u, v, d in g.edges.data():
                self.add_portal(renamed[u], renamed[v], symmetrical=True, **d)
        return self

    def become(self, g):
        """Erase all my nodes and edges. Replace them with a copy of the graph
        provided.

        Return myself.

        """
        self.clear()
        self.copy_from(g)
        return self

    def balanced_tree(self, r, h):
        return self.copy_from(nx.balanced_tree(r, h))

    def barbell_graph(self, m1, m2):
        return self.copy_from(nx.barbell_graph(m1, m2))

    def complete_graph(self, n):
        return self.copy_from(nx.complete_graph(n))

    def circular_ladder_graph(self, n):
        return self.copy_from(nx.circular_ladder_graph(n))

    def cycle_graph(self, n):
        return self.copy_from(nx.cycle_graph(n))

    def empty_graph(self, n):
        return self.copy_from(nx.empty_graph(n))

    def grid_2d_graph(self, m, n, periodic=False):
        return self.copy_from(nx.grid_2d_graph(m, n, periodic))

    def grid_2d_8graph(self, m, n):
        """Make a 2d graph that's connected 8 ways, enabling diagonal movement"""
        me = nx.Graph()
        node = me.node
        add_node = me.add_node
        add_edge = me.add_edge
        for i in range(m):
            for j in range(n):
                add_node((i, j))
                if i > 0:
                    add_edge((i, j), (i-1, j))
                    if j > 0:
                        add_edge((i, j), (i-1, j-1))
                if j > 0:
                    add_edge((i, j), (i, j-1))
                if (i - 1, j + 1) in node:
                    add_edge((i, j), (i-1, j+1))
        return self.copy_from(me)

    def grid_graph(self, dim, periodic=False):
        return self.copy_from(nx.grid_graph(dim, periodic))

    def ladder_graph(self, n):
        return self.copy_from(nx.ladder_graph(n))

    def lollipop_graph(self, m, n):
        return self.copy_from(nx.lollipop_graph(m, n))

    def path_graph(self, n):
        return self.copy_from(nx.path_graph(n))

    def star_graph(self, n):
        return self.copy_from(nx.star_graph(n))

    def wheel_graph(self, n):
        return self.copy_from(nx.wheel_graph(n))

    def fast_gnp_random_graph(self, n, p, seed=None):
        return self.copy_from(nx.fast_gnp_random_graph(
            n, p, seed, directed=True
        ))

    def gnp_random_graph(self, n, p, seed=None):
        return self.copy_from(nx.gnp_random_graph(n, p, seed, directed=True))

    def gnm_random_graph(self, n, m, seed=None):
        return self.copy_from(nx.gnm_random_graph(n, m, seed, directed=True))

    def erdos_renyi_graph(self, n, p, seed=None):
        return self.copy_from(nx.erdos_renyi_graph(n, p, seed, directed=True))

    def binomial_graph(self, n, p, seed=None):
        return self.erdos_renyi_graph(n, p, seed)

    def newman_watts_strogatz_graph(self, n, k, p, seed=None):
        return self.copy_from(nx.newman_watts_strogatz_graph(n, k, p, seed))

    def watts_strogatz_graph(self, n, k, p, seed=None):
        return self.copy_from(nx.watts_strogatz_graph(n, k, p, seed))

    def connected_watts_strogatz_graph(self, n, k, p, tries=100, seed=None):
        return self.copy_from(nx.connected_watts_strogatz_graph(
            n, k, p, tries, seed
        ))

    def random_regular_graph(self, d, n, seed=None):
        return self.copy_from(nx.random_regular_graph(d, n, seed))

    def barabasi_albert_graph(self, n, m, seed=None):
        return self.copy_from(nx.barabasi_albert_graph(n, m, seed))

    def powerlaw_cluster_graph(self, n, m, p, seed=None):
        return self.copy_from(nx.powerlaw_cluster_graph(n, m, p, seed))

    def duplication_divergence_graph(self, n, p, seed=None):
        return self.copy_from(nx.duplication_divergence_graph(n, p, seed))

    def random_lobster(self, n, p1, p2, seed=None):
        return self.copy_from(nx.random_lobster(n, p1, p2, seed))

    def random_shell_graph(self, constructor, seed=None):
        return self.copy_from(nx.random_shell_graph(constructor, seed))

    def random_powerlaw_tree(self, n, gamma=3, seed=None, tries=100):
        return self.copy_from(nx.random_powerlaw_tree(n, gamma, seed, tries))

    def configuration_model(self, deg_sequence, seed=None):
        return self.copy_from(nx.configuration_model(deg_sequence, seed=seed))

    def directed_configuration_model(
            self,
            in_degree_sequence,
            out_degree_sequence,
            seed=None
    ):
        return self.copy_from(nx.directed_configuration_model(
            in_degree_sequence,
            out_degree_sequence,
            seed=seed
        ))

    def expected_degree_graph(self, w, seed=None, selfloops=True):
        return self.copy_from(nx.expected_degree_graph(w, seed, selfloops))

    def havel_hakmi_graph(self, deg_sequence):
        return self.copy_from(nx.havel_hakimi_graph(deg_sequence))

    def directed_havel_hakmi_graph(
            self,
            in_degree_sequence,
            out_degree_sequence
    ):
        return self.copy_from(nx.directed_havel_hakmi_graph(
            in_degree_sequence,
            out_degree_sequence
        ))

    def degree_sequence_tree(self, deg_sequence):
        return self.copy_from(nx.degree_sequence_tree(deg_sequence))

    def random_degree_sequence_graph(self, sequence, seed=None, tries=10):
        return self.copy_from(nx.random_degree_sequence_graph(
            sequence, seed, tries
        ))

    def random_clustered_graph(self, joint_degree_sequence, seed=None):
        return self.copy_from(nx.random_clustered_graph(
            joint_degree_sequence, seed=seed
        ))

    def gn_graph(self, n, kernel=None, seed=None):
        return self.copy_from(nx.gn_graph(n, kernel, seed=seed))

    def gnr_graph(self, n, p, seed=None):
        return self.copy_from(nx.gnr_graph(n, p, seed=seed))

    def gnc_graph(self, n, seed=None):
        return self.copy_from(nx.gnc_graph(n, seed=seed))

    def scale_free_graph(
            self, n,
            alpha=0.41,
            beta=0.54,
            gamma=0.05,
            delta_in=0.2,
            delta_out=0,
            seed=None
    ):
        return self.copy_from(nx.scale_free_graph(
            n,
            alpha,
            beta,
            gamma,
            delta_in,
            delta_out,
            seed=seed
        ))

    def random_geometric_graph(self, n, radius, dim=2, pos=None):
        return self.copy_from(nx.random_geometric_graph(n, radius, dim, pos))

    def geographical_threshold_graph(
            self, n, theta,
            alpha=2,
            dim=2,
            pos=None,
            weight=None
    ):
        return self.copy_from(nx.geographical_threshold_graph(
            n, theta, alpha, dim, pos, weight
        ))

    def waxman_graph(
            self, n,
            alpha=0.4,
            beta=0.1,
            L=None,
            domain=(0, 0, 1, 1)
    ):
        return self.copy_from(nx.waxman_graph(
            n, alpha, beta, L, domain
        ))

    def navigable_small_world_graph(self, n, p=1, q=1, r=2, dim=2, seed=None):
        return self.copy_from(nx.navigable_small_world_graph(
            n, p, q, r, dim, seed
        ))

    def line_graph(self):
        lg = nx.line_graph(self)
        self.clear()
        return self.copy_from(lg)

    def ego_graph(
            self, n,
            radius=1,
            center=True,
            distance=None
    ):
        return self.become(nx.ego_graph(
            self, n,
            radius,
            center,
            False,
            distance
        ))

    def stochastic_graph(self, weight='weight'):
        nx.stochastic_graph(self, copy=False, weight=weight)
        return self

    def uniform_random_intersection_graph(self, n, m, p, seed=None):
        return self.copy_from(nx.uniform_random_intersection_graph(
            n, m, p, seed=seed
        ))

    def k_random_intersection_graph(self, n, m, k, seed=None):
        return self.copy_from(nx.k_random_intersection_graph(
            n, m, k, seed=seed
        ))

    def general_random_intersection_graph(self, n, m, p, seed=None):
        return self.copy_from(
            nx.general_random_intersection_graph(n, m, p, seed=seed)
        )

    def caveman_graph(self, l, k):
        return self.copy_from(nx.caveman_graph(l, k))

    def connected_caveman_graph(self, l, k):
        return self.copy_from(nx.connected_caveman_graph(l, k))

    def relaxed_caveman_graph(self, l, k, p, seed=None):
        return self.copy_from(nx.relaxed_caveman_graph(l, k, p, seed=seed))

    def random_partition_graph(self, sizes, p_in, p_out, seed=None):
        return self.copy_from(nx.random_partition_graph(
            sizes, p_in, p_out, seed=seed, directed=True
        ))

    def planted_partition_graph(self, l, k, p_in, p_out, seed=None):
        return self.copy_from(nx.planted_partition_graph(
            l, k, p_in, p_out, seed=seed, directed=True
        ))

    def gaussian_random_partition_graph(self, n, s, v, p_in, p_out, seed=None):
        return self.copy_from(nx.gaussian_random_partition_graph(
            n, s, v, p_in, p_out, seed=seed
        ))

    def _lookup_comparator(self, comparator):
        if callable(comparator):
            return comparator
        ops = {
            'ge': ge,
            'gt': gt,
            'le': le,
            'lt': lt,
            'eq': eq
        }
        if comparator in ops:
            return ops[comparator]
        return getattr(self.engine.function, comparator)

    def cull_nodes(self, stat, threshold=0.5, comparator=ge):
        """Delete nodes whose stat >= ``threshold`` (default 0.5).

        Optional argument ``comparator`` will replace >= as the test
        for whether to cull. You can use the name of a stored function.

        """
        comparator = self._lookup_comparator(comparator)
        dead = [
            name for name, node in self.node.items()
            if stat in node and comparator(node[stat], threshold)
        ]
        self.remove_nodes_from(dead)
        return self

    def cull_portals(self, stat, threshold=0.5, comparator=ge):
        """Delete portals whose stat >= ``threshold`` (default 0.5).

        Optional argument ``comparator`` will replace >= as the test
        for whether to cull. You can use the name of a stored function.

        """
        comparator = self._lookup_comparator(comparator)
        dead = []
        for u in self.portal:
            for v in self.portal[u]:
                if stat in self.portal[u][v] and comparator(
                        self.portal[u][v][stat], threshold
                ):
                    dead.append((u, v))
        self.remove_edges_from(dead)
        return self

    def cull_edges(self, stat, threshold=0.5, comparator=ge):
        """Delete edges whose stat >= ``threshold`` (default 0.5).

        Optional argument ``comparator`` will replace >= as the test
        for whether to cull. You can use the name of a stored function.

        """
        return self.cull_portals(stat, threshold, comparator)


class CharRuleMapping(RuleMapping):

    """Wraps one of a character's rulebooks so you can get its rules by name.

    You can access the rules in this either dictionary-style or as
    attributes. This is for convenience if you want to get at a rule's
    decorators, eg. to add an Action to the rule.

    Using this as a decorator will create a new rule, named for the
    decorated function, and using the decorated function as the
    initial Action.

    Using this like a dictionary will let you create new rules,
    appending them onto the underlying :class:`RuleBook`; replace one
    rule with another, where the new one will have the same index in
    the :class:`RuleBook` as the old one; and activate or deactivate
    rules. The name of a rule may be used in place of the actual rule,
    so long as the rule already exists.

    You can also set a rule active or inactive by setting it to
    ``True`` or ``False``, respectively. Inactive rules are still in
    the rulebook, but won't be followed.

    """

    def __init__(self, character, rulebook, booktyp):
        """Initialize as usual for the ``rulebook``, mostly.

        My ``character`` property will be the one passed in, and my
        ``_table`` will be the ``booktyp`` with ``"_rules"`` appended.

        """
        super().__init__(rulebook.engine, rulebook)
        self.character = character
        self._table = booktyp + "_rules"


class RuleFollower(BaseRuleFollower):

    """Mixin class. Has a rulebook, which you can get a RuleMapping into."""

    def _get_rule_mapping(self):
        return CharRuleMapping(
            self.character,
            self.rulebook,
            self._book
        )

    @abstractmethod
    def _get_rulebook_cache(self):
        pass

    def _get_rulebook_name(self):
        try:
            return self._get_rulebook_cache().retrieve(
                self.character.name, *self.engine.btt()
            )
        except KeyError:
            return self.character.name, self._book

    def _set_rulebook_name(self, n):
        branch, turn, tick = self.engine.nbtt()
        self.engine.query._set_rulebook_on_character(self._book, self.character.name, branch, turn, tick, n)
        self._get_rulebook_cache().store(self.character.name, branch, turn, tick, n)

    def __contains__(self, k):
        return self.engine._active_rules_cache.contains_key(
            self._get_rulebook_name(), *self.engine.btt()
        )


class SenseFuncWrap(object):

    """Wrapper for a sense function that looks it up in the code store if
    provided with its name, and prefills the first two arguments.

    """

    engine = getatt('character.engine')

    def __init__(self, character, fun):
        """Store the character and the function.

        Look up the function in the engine's ``sense`` function store,
        if needed.

        """
        if not callable(fun):
            raise TypeError("function is not callable")
        self.character = character
        if isinstance(fun, str):
            self.fun = self.engine.sense[fun]
        else:
            self.fun = fun

    def __call__(self, observed):
        """Call the function, prefilling the engine and observer arguments."""
        if isinstance(observed, str):
            observed = self.engine.character[observed]
        return self.fun(self.engine, self.character, Facade(observed))


class CharacterSense(object):

    """Mapping for when you've selected a sense for a character to use
    but haven't yet specified what character to look at

    """

    engine = getatt('container.engine')
    observer = getatt('container.character')

    def __init__(self, container, sensename):
        """Store the container and the name of the sense."""
        self.container = container
        self.sensename = sensename

    @property
    def func(self):
        """Return the function most recently associated with this sense."""
        fn = self.engine.query.sense_func_get(
            self.observer.name,
            self.sensename,
            *self.engine.btt()
        )
        if fn is not None:
            return SenseFuncWrap(self.observer, fn)

    def __call__(self, observed):
        """Call my sense function and make sure it returns the right type,
        then return that.

        """
        r = self.func(observed)
        if not (
                isinstance(r, Character) or
                isinstance(r, Facade)
        ):
            raise TypeError(
                "Sense function did not return a character-like object"
            )
        return r


class CharacterSenseMapping(MutableMappingUnwrapper, Signal):

    """Used to view other Characters as seen by one, via a particular sense."""

    # TODO: cache senses properly
    _book = "character"

    engine = getatt('character.engine')

    def __init__(self, character):
        """Store the character."""
        super().__init__()
        self.character = character

    def __iter__(self):
        """Iterate over active sense names."""
        yield from self.engine.query.sense_active_items(
            self.character.name, *self.engine.btt()
        )

    def __len__(self):
        """Count active senses."""
        n = 0
        for sense in iter(self):
            n += 1
        return n

    def __getitem__(self, k):
        """Get a :class:`CharacterSense` named ``k`` if it exists."""
        if not self.engine.query.sense_is_active(
                self.character.name,
                k,
                *self.engine.btt()
        ):
            raise KeyError("Sense isn't active or doesn't exist")
        return CharacterSense(self.character, k)

    def __setitem__(self, k, v):
        """Use the function for the sense from here on out."""
        if isinstance(v, str):
            funn = v
        else:
            funn = v.__name__
        if funn not in self.engine.sense:
            if not isinstance(v, Callable):
                raise TypeError("Not a function")
            self.engine.sense[funn] = v
        branch, turn, tick = self.engine.btt()
        # TODO: cache
        self.engine.query.sense_fun_set(
            self.character.name,
            k,
            branch,
            turn,
            tick,
            funn,
            True
        )
        self.send(self, key=k, val=v)

    def __delitem__(self, k):
        """Stop having the given sense."""
        branch, turn, tick = self.engine.btt()
        # TODO: cache
        self.engine.query.sense_set(
            self.character.name,
            k,
            branch,
            turn,
            tick,
            False
        )
        self.send(self, key=k, val=None)

    def __call__(self, fun, name=None):
        """Decorate the function so it's mine now."""
        if not isinstance(fun, Callable):
            raise TypeError(
                "I need a function here"
            )
        if name is None:
            name = fun.__name__
        self[name] = fun


class FacadeEntity(MutableMapping, Signal):
    def __init__(self, mapping, **kwargs):
        super().__init__()
        self.facade = mapping.facade
        self._real = mapping
        self._patch = kwargs
        self._masked = set()

    def __contains__(self, item):
        if item in self._masked:
            return False
        return item in self._patch or item in self._real

    def __iter__(self):
        seen = set()
        for k in self._real:
            if k not in self._masked:
                yield k
            seen.add(k)
        for k in self._patch:
            if (
                    k not in self._masked and
                    k not in seen
            ):
                yield k

    def __len__(self):
        n = 0
        for k in self:
            n += 1
        return n

    def __getitem__(self, k):
        if k in self._masked:
            raise KeyError("{} has been masked.".format(k))
        if k in self._patch:
            return self._patch[k]
        ret = self._real[k]
        if hasattr(ret, 'unwrap'):  # a wrapped mutable object from the allegedb.wrap module
            ret = ret.unwrap()
            self._patch[k] = ret  # changes will be reflected in the facade but not the original
        return ret

    def __setitem__(self, k, v):
        if k == 'name':
            raise TypeError("Can't change names")
        self._masked.discard(k)
        self._patch[k] = v
        self.send(self, key=k, val=v)

    def __delitem__(self, k):
        self._masked.add(k)
        self.send(self, key=k, val=None)


class FacadePlace(FacadeEntity):

    """Lightweight analogue of Place for Facade use."""

    @property
    def name(self):
        return self['name']

    def contents(self):
        for thing in self.facade.thing.values():
            if thing.container is self:
                yield thing

    def __init__(self, mapping, real_or_name, **kwargs):
        super().__init__(mapping, **kwargs)
        if isinstance(real_or_name, Place) or \
           isinstance(real_or_name, FacadePlace):
            self._real = real_or_name
        else:
            self._real = {'name': real_or_name}


class FacadeThing(FacadeEntity):
    @property
    def name(self):
        return self._real['name']

    def __init__(self, mapping, real_or_name, **kwargs):
        location = kwargs.pop('location', None)
        super().__init__(mapping, **kwargs)
        if location is None and not (
                isinstance(real_or_name, Thing) or
                isinstance(real_or_name, FacadeThing)
        ):
            raise TypeError(
                "FacadeThing needs to wrap a real Thing or another "
                "FacadeThing, or have a location of its own."
            )
        self._real = {
            'name': real_or_name.name if hasattr(real_or_name, 'name')
            else real_or_name,
            'location': location
        }

    @property
    def location(self):
        try:
            return self.facade.node[self['location']]
        except KeyError:
            return None


class FacadePortal(FacadeEntity):
    """Lightweight analogue of Portal for Facade use."""

    def __init__(self, mapping, other, **kwargs):
        super().__init__(mapping, **kwargs)
        if hasattr(mapping, 'orig'):
            self.orig = mapping.orig
            self.dest = other
        else:
            self.dest = mapping.dest
            self.orig = other
        try:
            self._real = self.facade.character.portal[self.orig][self.dest]
        except (KeyError, AttributeError):
            self._real = {}

    def __setitem__(self, k, v):
        if k in ('origin', 'destination'):
            raise TypeError("Portals have fixed origin and destination")
        super().__setitem__(k, v)

    @property
    def origin(self):
        return self.facade.node[self.orig]

    @property
    def destination(self):
        return self.facade.node[self.dest]


class FacadeEntityMapping(MutableMappingUnwrapper, Signal):

    """Mapping that contains entities in a Facade.

    All the entities are of the same type, ``facadecls``, possibly
    being distorted views of entities of the type ``innercls``.

    """
    def _make(self, k, v):
        return self.facadecls(self, k, **v)

    engine = getatt('facade.engine')

    def __init__(self, facade):
        """Store the facade."""
        super().__init__()
        self.facade = facade
        self._patch = {}
        self._masked = set()

    def __contains__(self, k):
        return (
            k not in self._masked and (
                k in self._patch or
                k in self._get_inner_map()
            )
        )

    def __iter__(self):
        seen = set()
        for k in self._get_inner_map():
            if k not in self._masked:
                yield k
            seen.add(k)
        for k in self._patch:
            if k not in seen:
                yield k

    def __len__(self):
        n = 0
        for k in self:
            n += 1
        return n

    def __getitem__(self, k):
        if k not in self:
            raise KeyError
        if k not in self._patch:
            self._patch[k] = self.facadecls(self, k, **self._get_inner_map()[k])
        return self._patch[k]

    def __setitem__(self, k, v):
        if not isinstance(v, self.facadecls):
            v = self._make(k, v)
        self._masked.discard(k)
        self._patch[k] = v
        self.send(self, key=k, val=v)

    def __delitem__(self, k):
        self._masked.add(k)
        self.send(self, key=k, val=None)


class FacadePortalSuccessors(FacadeEntityMapping):
    facadecls = FacadePortal
    innercls = Portal

    def __init__(self, facade, origname):
        super().__init__(facade)
        self.orig = origname

    def _make(self, k, v):
        return self.facadecls(self, k, **v)

    def _get_inner_map(self):
        try:
            return self.facade.character.portal[self.orig]
        except AttributeError:
            return {}


class FacadePortalPredecessors(FacadeEntityMapping):
    facadecls = FacadePortal
    innercls = Portal

    def __init__(self, facade, destname):
        super().__init__(facade)
        self.dest = destname

    def _make(self, k, v):
        return self.facadecls(self.facade.portal[k], self.dest, v)

    def _get_inner_map(self):
        try:
            return self.facade.character.preportal[self._destname]
        except AttributeError:
            return {}


class FacadePortalMapping(FacadeEntityMapping):
    def __getitem__(self, node):
        if node in self._masked:
            raise KeyError("Node {} is in the inner Character, but has been masked".format(node))
        if node not in self:
            raise KeyError("No such node: {}".format(node))
        if node not in self._patch:
            self._patch[node] = self.cls(self.facade, node)
        return self._patch[node]

    def __setitem__(self, node, value):
        self._masked.discard(node)
        v = self.cls(self.facade, node)
        v.update(value)
        self._patch[node] = v

    def __delitem__(self, node):
        self._masked.add(node)


class Facade(AbstractCharacter, nx.DiGraph):
    engine = getatt('character.engine')

    def __getstate__(self):
        ports = {}
        for o in self.portal:
            if o not in ports:
                ports[o] = {}
            for d in self.portal[o]:
                ports[o][d] = dict(self.portal[o][d])
        things = {k: dict(v) for (k, v) in self.thing.items()}
        places = {k: dict(v) for (k, v) in self.place.items()}
        return things, places, ports

    def __setstate__(self, state):
        self.character = None
        self.thing, self.place, self.portal = state

    def add_places_from(self, seq, **attrs):
        for place in seq:
            self.add_place(place, **attrs)

    def add_things_from(self, seq, **attrs):
        for thing in seq:
            self.add_thing(thing, **attrs)

    def thing2place(self, name):
        self.place[name] = self.thing.pop(name)

    def place2thing(self, name, location):
        it = self.place.pop(name)
        it['location'] = location
        self.thing[name] = it

    def add_portals_from(self, seq, **attrs):
        for it in seq:
            self.add_portal(*it, **attrs)

    def del_avatar(self, a, b=None):
        raise NotImplementedError("Facades don't have avatars")

    def add_place(self, name, **kwargs):
        self.place[name] = kwargs

    def add_node(self, name, **kwargs):
        self.place[name] = kwargs

    def add_thing(self, name, **kwargs):
        self.thing[name] = kwargs

    def add_portal(self, orig, dest, symmetrical=False, **kwargs):
        self.portal[orig][dest] = kwargs
        if symmetrical:
            mirror = dict(kwargs)
            mirror['is_mirror'] = True
            self.portal[dest][orig] = mirror

    def add_edge(self, orig, dest, **kwargs):
        self.add_portal(orig, dest, **kwargs)

    def add_avatar(self, a, b=None):
        raise NotImplementedError("Facades don't have avatars")

    def __init__(self, character=None):
        """Store the character."""
        self.character = character
        self.graph = self.StatMapping(self)

    class ThingMapping(FacadeEntityMapping):
        facadecls = FacadeThing
        innercls = Thing

        def _get_inner_map(self):
            try:
                return self.facade.character.thing
            except AttributeError:
                return {}

    class PlaceMapping(FacadeEntityMapping):
        facadecls = FacadePlace
        innercls = Place

        def _get_inner_map(self):
            try:
                return self.facade.character.place
            except AttributeError:
                return {}

    def ThingPlaceMapping(self, *args):
        return CompositeDict(self.thing, self.place)

    class PortalSuccessorsMapping(FacadePortalMapping):
        cls = FacadePortalSuccessors

        def __contains__(self, item):
            return item in self.facade.node

        def _get_inner_map(self):
            try:
                return self.facade.character.portal
            except AttributeError:
                return {}

    class PortalPredecessorsMapping(FacadePortalMapping):
        cls = FacadePortalPredecessors

        def __contains__(self, item):
            return item in self.facade.node

        def _get_inner_map(self):
            try:
                return self.facade.character.preportal
            except AttributeError:
                return {}

    class StatMapping(MutableMappingUnwrapper, Signal):
        def __init__(self, facade):
            super().__init__()
            self.facade = facade
            self._patch = {}
            self._masked = set()

        def __iter__(self):
            seen = set()
            if hasattr(self.facade.character, 'graph'):
                for k in self.facade.character.graph:
                    if k not in self._masked:
                        yield k
                    seen.add(k)
            for k in self._patch:
                if k not in seen:
                    yield k

        def __len__(self):
            n = 0
            for k in self:
                n += 1
            return n

        def __contains__(self, k):
            if k in self._masked:
                return False
            if hasattr(self.facade.character, 'graph') and k in self.facade.character.graph:
                return True
            return k in self._patch

        def __getitem__(self, k):
            if k in self._masked:
                raise KeyError("masked")
            if k not in self._patch and hasattr(self.facade.character, 'graph'):
                ret = self.facade.character.graph[k]
                if not hasattr(ret, 'unwrap'):
                    return ret
                self._patch[k] = ret.unwrap()
            return self._patch[k]

        def __setitem__(self, k, v):
            self._masked.discard(k)
            self._patch[k] = v
            self.send(self, key=k, val=v)

        def __delitem__(self, k):
            self._masked.add(k)
            self.send(self, key=k, val=None)


class Character(DiGraph, AbstractCharacter, RuleFollower):
    """A graph that follows game rules and has a containment hierarchy.

    Nodes in a Character are subcategorized into Things and
    Places. Things have locations, and those locations may be Places
    or other Things.

    Characters may have avatars in other Characters. These are just
    nodes. You can apply rules to a Character's avatars, and thus to
    any collection of nodes you want, perhaps in many different
    Characters. But you may want a Character to have exactly one
    avatar, representing their location in physical space -- the
    Character named 'physical'. So when a Character has only one
    avatar, you can treat the ``avatar`` property as an alias of the
    avatar.


    """
    _book = "character"

    @property
    def character(self):
        return self

    def _get_rulebook_cache(self):
        return self.engine._characters_rulebooks_cache

    def __repr__(self):
        return "{}.character[{}]".format(repr(self.engine), repr(self.name))

    def __init__(self, engine, name, data=None, *, init_rulebooks=True, **attr):
        """Store engine and name, and set up mappings for Thing, Place, and
        Portal

        """
        from allegedb.cache import FuturistWindowDict, PickyDefaultDict
        super().__init__(engine, name, data, **attr)
        self._avatars_cache = PickyDefaultDict(FuturistWindowDict)
        if not init_rulebooks:
            return
        cachemap = {
            'character': engine._characters_rulebooks_cache,
            'avatar': engine._avatars_rulebooks_cache,
            'character_thing': engine._characters_things_rulebooks_cache,
            'character_place': engine._characters_places_rulebooks_cache,
            'character_portal': engine._characters_portals_rulebooks_cache
        }
        for rulebook, cache in cachemap.items():
            branch, turn, tick = engine.nbtt()
            rulebook_or_name = attr.get(rulebook, (name, rulebook))
            rulebook_name = getattr(rulebook_or_name, 'name', rulebook_or_name)
            engine.query._set_rulebook_on_character(rulebook, name, branch, turn, tick, rulebook_name)
            cache.store((name, rulebook), branch, turn, tick, rulebook_name)

    class ThingMapping(MutableMappingUnwrapper, RuleFollower, Signal):
        """:class:`Thing` objects that are in a :class:`Character`"""
        _book = "character_thing"

        engine = getatt('character.engine')
        name = getatt('character.name')

        def _get_rulebook_cache(self):
            return self.engine._characters_things_rulebooks_cache

        def __init__(self, character):
            """Store the character and initialize cache."""
            super().__init__()
            self.character = character

        def __iter__(self):
            cache = self.engine._things_cache
            char = self.name
            branch, turn, tick = self.engine.btt()
            for key in cache.iter_keys(char, branch, turn, tick):
                try:
                    if cache.retrieve(char, key, branch, turn, tick) is not None:
                        yield key
                except KeyError:
                    continue

        def __contains__(self, thing):
            args = self.character.name, thing, *self.engine.btt()
            cache = self.engine._things_cache
            return cache.contains_key(*args) and cache.retrieve(*args) is not None

        def __len__(self):
            return self.engine._things_cache.count_keys(
                self.character.name, *self.engine.btt()
            )

        def __getitem__(self, thing):
            if thing not in self:
                raise KeyError("No such thing: {}".format(thing))
            return self._make_thing(thing)

        def _make_thing(self, thing, val=None):
            cache = self.engine._node_objs
            if isinstance(val, Thing):
                th = cache[self.name, thing] = val
            elif (self.name, thing) in cache:
                th = cache[(self.name, thing)]
                if type(th) is not Thing:
                    th = cache[self.name, thing] = Thing(self.character, thing)
            else:
                th = cache[(self.name, thing)] = Thing(self.character, thing)
            return th

        def __setitem__(self, thing, val):
            if not isinstance(val, Mapping):
                raise TypeError('Things are made from Mappings')
            if 'location' not in val:
                raise ValueError('Thing needs location')
            created = thing not in self
            self.engine._exist_node(self.character.name, thing)
            self.engine._set_thing_loc(
                self.character.name,
                thing,
                val['location']
            )
            th = self._make_thing(thing, val)
            th.clear()
            th.update(val)
            if created:
                self.send(self, thing_name=thing, exists=True)

        def __delitem__(self, thing):
            self[thing].delete()
            self.send(self, thing_name=thing, exists=False)

        def __repr__(self):
            return "{}.character[{}].thing".format(repr(self.engine), repr(self.name))

    class PlaceMapping(MutableMappingUnwrapper, RuleFollower, Signal):
        """:class:`Place` objects that are in a :class:`Character`"""
        _book = "character_place"

        engine = getatt('character.engine')
        name = getatt('character.name')

        def _get_rulebook_cache(self):
            return self.engine._characters_places_rulebooks_cache

        def __init__(self, character):
            """Store the character."""
            super().__init__()
            self.character = character

        def __iter__(self):
            for node in self.engine._nodes_cache.iter_entities(
                    self.character.name, *self.engine.btt()
            ):
                if not self.engine._things_cache.contains_entity(
                        self.character.name, node, *self.engine.btt()
                ):
                    yield node

        def __len__(self):
            return self.engine._nodes_cache.count_entities(
                self.character.name, *self.engine.btt()
            ) - self.engine._things_cache.count_entities(
                self.character.name, *self.engine.btt()
            )

        def __contains__(self, place):
            # TODO: maybe a special cache just for places and not just
            # nodes in general
            return (
                self.engine._nodes_cache.contains_entity(
                    self.character.name, place, *self.engine.btt()
                ) and not self.engine._things_cache.contains_entity(
                    self.character.name, place, *self.engine.btt()
                )
            )

        def __getitem__(self, place):
            if place not in self:
                raise KeyError("No such place: {}".format(place))
            cache = self.engine._node_objs
            if (self.name, place) not in cache or not isinstance(
                    cache[(self.name, place)], Place
            ):
                ret = cache[(self.name, place)] = Place(self.character, place)
                return ret
            return cache[(self.name, place)]

        def __setitem__(self, place, v):
            pl = self.engine._get_node(self.character, place)
            if not isinstance(pl, Place):
                raise KeyError("{} is not a place".format(place))
            pl.clear()
            pl.update(v)
            self.send(self, key=place, val=v)

        def __delitem__(self, place):
            self[place].delete()

        def __repr__(self):
            return "{}.character[{}].place".format(repr(self.engine), repr(self.name))

    class ThingPlaceMapping(GraphNodeMapping, Signal):
        """GraphNodeMapping but for Place and Thing"""
        _book = "character_node"

        character = getatt('graph')
        engine =  getatt('db')
        name = getatt('character.name')

        def __init__(self, character):
            """Store the character."""
            super().__init__(character)
            Signal.__init__(self)

        def __contains__(self, k):
            return self.engine._node_exists(self.character.name, k)

        def __getitem__(self, k):
            if k not in self:
                raise KeyError
            return self.engine._get_node(self.character, k)

        def __setitem__(self, k, v):
            self.character.place[k] = v

        def __delitem__(self, k):
            if k not in self:
                raise KeyError
            if self.engine._is_thing(
                self.character.name, k
            ):
                del self.character.thing[k]
            else:
                del self.character.place[k]
    node_map_cls = ThingPlaceMapping

    class PortalSuccessorsMapping(GraphSuccessorsMapping, RuleFollower):
        """Mapping of nodes that have at least one outgoing edge.

        Maps them to another mapping, keyed by the destination nodes,
        which maps to Portal objects.

        """
        _book = "character_portal"

        character = getatt('graph')
        engine = getatt('graph.engine')

        def _get_rulebook_cache(self):
            return self.engine._characters_portals_rulebooks_cache

        def __getitem__(self, orig):
            if self.engine._node_exists(
                    self.graph.name,
                    orig
            ):
                if orig not in self._cache:
                    self._cache[orig] = self.Successors(self, orig)
                return self._cache[orig]
            raise KeyError("No such node")

        def __setitem__(self, orig, val):
            if orig not in self._cache:
                self._cache[orig] = self.Successors(self, orig)
            sucs = self._cache[orig]
            sucs.clear()
            sucs.update(val)
            self.send(self, key=orig, val=sucs)

        def __delitem__(self, orig):
            super().__delitem__(orig)
            self.send(self, key=orig, val=None)

        class Successors(GraphSuccessorsMapping.Successors):
            """Mapping for possible destinations from some node."""

            engine = getatt('graph.engine')

            @staticmethod
            def send(self, **kwargs):
                """Call all listeners to ``dest`` and to my ``orig``."""
                super().send(self, **kwargs)
                self.container.send(self, **kwargs)

            def __getitem__(self, dest):
                if dest in self:
                    return self.engine._get_edge(self.graph, self.orig, dest, 0)
                raise KeyError("No such portal: {}->{}".format(
                    self.orig, dest
                ))

            def __setitem__(self, dest, value):
                self.engine._exist_edge(
                    self.graph.name,
                    self.orig,
                    dest
                )
                p = self.engine._get_edge(self.graph, self.orig, dest, 0)
                p.clear()
                p.update(value)
                self.send(self, key=dest, val=p)

            def __delitem__(self, dest):
                if dest not in self:
                    raise KeyError("No portal to {}".format(dest))
                self[dest].delete()
    adj_cls = PortalSuccessorsMapping

    class PortalPredecessorsMapping(
            DiGraphPredecessorsMapping,
            RuleFollower
    ):
        """Mapping of nodes that have at least one incoming edge.

        Maps to another mapping keyed by the origin nodes, which maps to
        Portal objects.

        """
        _book = "character_portal"

        def _get_rulebook_cache(self):
            return self.engine._characters_portals_rulebooks_cache

        class Predecessors(DiGraphPredecessorsMapping.Predecessors):
            """Mapping of possible origins from some destination."""
            def __setitem__(self, orig, value):
                key = (self.graph.name, orig, self.dest)
                if key not in self.db._portal_objs:
                    self.db._portal_objs[key] = Portal(
                        self.graph,
                        orig,
                        self.dest
                    )
                p = self.db._portal_objs[key]
                p.clear()
                p.update(value)
                p.engine._exist_edge(self.graph.name, self.dest, orig)
    pred_cls = PortalPredecessorsMapping

    class AvatarGraphMapping(Mapping, RuleFollower):
        """A mapping of other characters in which one has an avatar.

        Maps to a mapping of the avatars themselves, unless there's
        only one other character you have avatars in, in which case
        this maps to those.

        If you have only one avatar anywhere, you can pretend this
        is that entity.

        """
        _book = "avatar"

        engine = getatt('character.engine')
        name = getatt('character.name')

        def _get_rulebook_cache(self):
            return self.engine._avatars_rulebooks_cache

        def __init__(self, char):
            """Remember my character."""
            self.character = char
            self._char_av_cache = {}

        def __call__(self, av):
            """Add the avatar. It must be an instance of Place or Thing."""
            if av.__class__ not in (Place, Thing):
                raise TypeError("Only Things and Places may be avatars")
            self.character.add_avatar(av.name, av.character.name)

        def __iter__(self):
            """Iterate over every avatar graph that has at least one avatar node
            in it presently

            """
            return iter(self.engine._avatarness_cache.get_char_graphs(
                self.character.name, *self.engine.btt()
            ))

        def __contains__(self, k):
            return k in self.engine._avatarness_cache.get_char_graphs(
                self.character.name, *self.engine.btt()
            )

        def __len__(self):
            """Number of graphs in which I have an avatar."""
            return len(self.engine._avatarness_cache.get_char_graphs(
                self.character.name, *self.engine.btt()
            ))

        def _get_char_av_cache(self, g):
            if g not in self:
                raise KeyError
            if g not in self._char_av_cache:
                self._char_av_cache[g] = self.CharacterAvatarMapping(self, g)
            return self._char_av_cache[g]

        def __getitem__(self, g):
            return self._get_char_av_cache(g)

        @property
        def node(self):
            """If I have avatars in only one graph, return a map of them.

            Otherwise, raise AttributeError.

            """
            try:
                return self._get_char_av_cache(
                    self.engine._avatarness_cache.get_char_only_graph(
                        self.character.name, *self.engine.btt()
                    )
                )
            except KeyError:
                raise AttributeError(
                    "I have no avatar, or I have avatars in many graphs"
                )

        @property
        def only(self):
            """If I have only one avatar, return it.

            Otherwise, raise AttributeError.

            """
            try:
                charn, noden = self.engine._avatarness_cache.get_char_only_av(
                    self.character.name, *self.engine.btt()
                )
                return self.engine._get_node(self.engine.character[charn], noden)
            except KeyError:
                raise AttributeError(
                    "I have no avatar, or more than one avatar"
                )

        class CharacterAvatarMapping(Mapping):
            """Mapping of avatars of one Character in another Character."""
            def __init__(self, outer, graphn):
                """Store the character and the name of the "graph", ie. the other
                character.

                """
                self.character = outer.character
                self.engine = outer.engine
                self.name = outer.name
                self.graph = graphn

            def __iter__(self):
                """Iterate over the names of all the presently existing nodes in the
                graph that are avatars of the character

                """
                return iter(self.engine._avatarness_cache.get_char_graph_avs(
                    self.name, self.graph, *self.engine.btt()
                ))

            def __contains__(self, av):
                return av in self.engine._avatarness_cache.get_char_graph_avs(
                    self.name, self.graph, *self.engine.btt()
                )

            def __len__(self):
                """Number of presently existing nodes in the graph that are avatars of
                the character"""
                return len(self.engine._avatarness_cache.get_char_graph_avs(
                    self.name, self.graph, *self.engine.btt()
                ))

            def __getitem__(self, av):
                if av in self:
                    return self.engine._get_node(self.engine.character[self.graph], av)
                raise KeyError("No avatar: {}".format(av))

            @property
            def only(self):
                mykey = singleton_get(self.keys())
                if mykey is None:
                    raise AttributeError("No avatar, or more than one")
                return self.engine._get_node(self.engine.character[self.graph], mykey)

            def __setitem__(self, k, v):
                mykey = singleton_get(self.keys())
                if mykey is None:
                    raise AmbiguousAvatarError(
                        "More than one avatar in {}; "
                        "be more specific to set the stats of one.".format(
                            self.graph
                        )
                    )
                self.engine._get_node(self.graph, mykey)[k] = v

            def __repr__(self):
                return "{}.character[{}].avatar".format(repr(self.engine), repr(self.name))

    def facade(self):
        return Facade(self)

    def add_place(self, n, **kwargs):
        super().add_node(n, **kwargs)

    def add_places_from(self, seq, **attrs):
        """Take a series of place names and add the lot."""
        super().add_nodes_from(seq, **attrs)

    def add_thing(self, name, location, **kwargs):
        """Create a Thing, set its location,
        and set its initial attributes from the keyword arguments (if
        any).

        """
        if name in self.thing:
            raise WorldIntegrityError(
                "Already have a Thing named {}".format(name)
            )
        self.add_node(name, **kwargs)
        if isinstance(location, Node):
            location = location.name
        self.place2thing(name, location,)

    def add_things_from(self, seq, **attrs):
        for tup in seq:
            name = tup[0]
            location = tup[1]
            kwargs = tup[2] if len(tup) > 2 else attrs
            self.add_thing(name, location, **kwargs)

    def place2thing(self, name, location):
        """Turn a Place into a Thing with the given location.
        
        It will keep all its attached Portals.

        """
        self.engine._set_thing_loc(
            self.name, name, location
        )
        if (self.name, name) in self.engine._node_objs:
            del self.engine._node_objs[self.name, name]

    def thing2place(self, name):
        """Unset a Thing's location, and thus turn it into a Place."""
        self.engine._set_thing_loc(
            self.name, name, None
        )
        if (self.name, name) in self.engine._node_objs:
            del self.engine._node_objs[self.name, name]

    def add_portal(self, origin, destination, symmetrical=False, **kwargs):
        """Connect the origin to the destination with a :class:`Portal`.

        Keyword arguments are the :class:`Portal`'s
        attributes. Exception: if keyword ``symmetrical`` == ``True``,
        a mirror-:class:`Portal` will be placed in the opposite
        direction between the same nodes. It will always appear to
        have the placed :class:`Portal`'s stats, and any change to the
        mirror :class:`Portal`'s stats will affect the placed
        :class:`Portal`.

        """
        if isinstance(origin, Node):
            origin = origin.name
        if isinstance(destination, Node):
            destination = destination.name
        super().add_edge(origin, destination, **kwargs)
        if symmetrical:
            self.add_portal(destination, origin, is_mirror=True)

    def new_portal(self, origin, destination, symmetrical=False, **kwargs):
        if isinstance(origin, Node):
            origin = origin.name
        if isinstance(destination, Node):
            destination = destination.name
        self.add_portal(origin, destination, symmetrical, **kwargs)
        return self.engine._get_edge(self, origin, destination, 0)

    def add_portals_from(self, seq, symmetrical=False):
        """Take a sequence of (origin, destination) pairs and make a
        :class:`Portal` for each.

        Actually, triples are acceptable too, in which case the third
        item is a dictionary of stats for the new :class:`Portal`.

        If optional argument ``symmetrical`` is set to ``True``, all
        the :class:`Portal` instances will have a mirror portal going
        in the opposite direction, which will always have the same
        stats.

        """
        for tup in seq:
            orig = tup[0]
            dest = tup[1]
            kwargs = tup[2] if len(tup) > 2 else {}
            if symmetrical:
                kwargs['symmetrical'] = True
            self.add_portal(orig, dest, **kwargs)

    def add_avatar(self, a, b=None):
        """Start keeping track of a :class:`Thing` or :class:`Place` in a
        different :class:`Character`.

        """
        if b is None:
            if not (
                    isinstance(a, Place) or
                    isinstance(a, Thing)
            ):
                raise TypeError(
                    'when called with one argument, '
                    'it must be a place or thing'
                )
            g = a.character.name
            n = a.name
        else:
            if isinstance(a, Character):
                g = a.name
            elif not isinstance(a, str):
                raise TypeError(
                    'when called with two arguments, '
                    'the first is a character or its name'
                )
            else:
                g = a
            if isinstance(b, Place) or isinstance(b, Thing):
                n = b.name
            elif not isinstance(b, str):
                raise TypeError(
                    'when called with two arguments, '
                    'the second is a thing/place or its name'
                )
            else:
                n = b
        # This will create the node if it doesn't exist. Otherwise
        # it's redundant but harmless.
        self.engine._exist_node(g, n)
        # Declare that the node is my avatar
        branch, turn, tick = self.engine.nbtt()
        self.engine._remember_avatarness(self.name, g, n, branch=branch, turn=turn, tick=tick)

    def del_avatar(self, a, b=None):
        """This is no longer my avatar, though it still exists on its own."""
        if b is None:
            if not isinstance(a, Node):
                raise TypeError(
                    "In single argument form, "
                    "del_avatar requires a Node object "
                    "(Thing or Place)."
                )
            g = a.character.name
            n = a.name
        else:
            g = a.name if isinstance(a, Character) else a
            n = b.name if isinstance(b, Node) else b
        self.engine._remember_avatarness(
            self.character.name, g, n, False
        )

    def portals(self):
        """Iterate over all portals."""
        char = self.character
        make_edge = self.engine._get_edge
        for (o, d) in self.engine._edges_cache.iter_keys(
                self.character.name, *self.engine.btt()
        ):
            yield make_edge(char, o, d)

    def avatars(self):
        """Iterate over all my avatars, regardless of what character they are
        in.

        """
        charname = self.character.name
        branch, turn, tick = self.engine.btt()
        charmap = self.engine.character
        avit = self.engine._avatarness_cache.iter_entities
        makenode = self.engine._get_node
        for graph in avit(
                charname, branch, turn, tick
        ):
            for node in avit(
                    charname, graph, branch, turn, tick
            ):
                try:
                    yield makenode(charmap[graph], node)
                except KeyError:
                    continue



# This file is part of LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import cherrypy
from argparse import ArgumentParser
from . import LiSEHandleWebService

parser = ArgumentParser()
parser.add_argument('world', action='store', required=True)
parser.add_argument('-c', '--code', action='store')
args = parser.parse_args()
conf = {
    '/': {
        'request.dispatch': cherrypy.dispatch.MethodDispatcher(),
        'tools.sessions.on': True,
        'tools.response_headers.on': True,
        'tools.response_headers.headers': [('Content-Type', 'application/json')],
        'tools.encode.on': True,
        'tools.encode.encoding': 'utf-8'
    }
}
cherrypy.quickstart(LiSEHandleWebService(args.world, args.code), '/', conf)



# This file is part of LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""A base class for nodes that can be in a character.

Every actual node that you're meant to use will be a place or
thing. This module is for what they have in common.

"""
from collections import Mapping, ValuesView

from networkx import shortest_path, shortest_path_length

import allegedb.graph
from allegedb.cache import HistoryError

from .util import getatt
from .query import StatusAlias
from . import rule
from .exc import AmbiguousUserError


class RuleMapping(rule.RuleMapping):
    """Version of :class:`LiSE.rule.RuleMapping` that works more easily
    with a node.

    """
    def __init__(self, node):
        """Initialize with node's engine, character, and rulebook."""
        super().__init__(node.engine, node.rulebook)
        self.node = node


class UserMapping(Mapping):
    """A mapping of the characters that have a particular node as an avatar.

    Getting characters from here isn't any better than getting them from
    the engine direct, but with this you can do things like use the
    .get() method to get a character if it's a user and otherwise
    get something else; or test whether the character's name is in
    the keys; and so on.

    """
    __slots__ = ['node']

    def __init__(self, node):
        """Store the node"""
        self.node = node

    engine = getatt('node.engine')

    def _user_names(self):
        cache = self.engine._avatarness_cache.user_order
        if self.node.character.name not in cache or \
           self.node.name not in cache[self.node.character.name]:
            return
        cache = cache[self.node.character.name][self.node.name]
        seen = set()
        for user in cache:
            if user in seen:
                continue
            for (branch, turn, tick) in self.node.engine._iter_parent_btt():
                if branch in cache[user]:
                    branchd = cache[user][branch]
                    try:
                        if turn in branchd:
                            if branchd[turn].get(tick, False):
                                yield user
                        elif branchd.rev_gettable(turn):
                            turnd = branchd[turn]
                            if turnd[turnd.end]:
                                yield user
                        seen.add(user)
                        break
                    except HistoryError as ex:
                        if ex.deleted:
                            break

    def __iter__(self):
        yield from self._user_names()

    def __len__(self):
        n = 0
        for user in self._user_names():
            n += 1
        return n

    def __contains__(self, item):
        if item in self.engine.character:
            item = self.engine.character[item]
        if hasattr(item, 'avatar'):
            charn = self.node.character.name
            nn = self.node.name
            return charn in item.avatar and nn in item.avatar[charn]
        return False

    def __getitem__(self, k):
        if k not in self:
            raise KeyError("{} not used by {}".format(
                self.node.name, k
            ))
        return self.engine.character[k]


class NodeContentValues(ValuesView):
    def __iter__(self):
        node = self._mapping.node
        nodem = node.character.node
        try:
            for name in node.engine._node_contents_cache.retrieve(
                    node.character.name, node.name, *node.engine.btt()
            ):
                yield nodem[name]
        except KeyError:
            return

    def __contains__(self, item):
        return item.location == self._mapping.node


class NodeContent(Mapping):
    __slots__ = ('node',)

    def __init__(self, node):
        self.node = node

    def __iter__(self):
        try:
            yield from self.node.engine._node_contents_cache.retrieve(
                self.node.character.name, self.node.name, *self.node.engine.btt()
            )
        except KeyError:
            return

    def __len__(self):
        try:
            return len(self.node.engine._node_contents_cache.retrieve(
                self.node.character.name, self.node.name, *self.node.engine.btt()
            ))
        except KeyError:
            return 0

    def __contains__(self, item):
        try:
            return self.node.character.thing[item].location == self.node
        except KeyError:
            return False

    def __getitem__(self, item):
        if item not in self:
            raise KeyError
        return self.node.character.thing[item]

    def values(self):
        return NodeContentValues(self)


class DestsValues(ValuesView):
    def __contains__(self, item):
        return item.origin == self._mapping.node


class Dests(Mapping):
    __slots__ = ('node',)

    def __init__(self, node):
        self.node = node

    def __iter__(self):
        yield from self.node.engine._edges_cache.iter_successors(
            self.node.character.name, self.node.name, *self.node.engine.btt()
        )

    def __len__(self):
        return self.node.engine._edges_cache.count_successors(
            self.node.character.name, self.node.name, *self.node.engine.btt()
        )

    def __contains__(self, item):
        return self.node.engine._edges_cache.has_successor(
            self.node.character.name, self.node.name, item, *self.node.engine.btt()
        )

    def __getitem__(self, item):
        if item not in self:
            raise KeyError
        return self.node.character.portal[self.node.name][item]

    def values(self):
        return DestsValues(self)


class OrigsValues(ValuesView):
    def __contains__(self, item):
        return item.destination == self._mapping.node


class Origs(Mapping):
    __slots__ = ('node',)

    def __init__(self, node):
        self.node = node

    def __iter__(self):
        return self.node.engine._edges_cache.iter_predecessors(
            self.node.character.name, self.node.name, *self.node.engine.btt()
        )

    def __contains__(self, item):
        return self.node.engine._edges_cache.has_predecessor(
            self.node.character.name, self.node.name, item, *self.node.engine.btt()
        )

    def __len__(self):
        return self.node.engine._edges_cache.count_predecessors(
            self.node.character.name, self.node.name, *self.node.engine.btt()
        )

    def __getitem__(self, item):
        if item not in self:
            raise KeyError
        return self.node.character.portal[item][self.node.name]

    def values(self):
        return OrigsValues(self)


class UserDescriptor:
    """Give a node's user if there's only one

    If there are many users, but one of them has the same name as this node, give that one.

    Otherwise, raise AmbiguousUserError.

    """
    usermapping = UserMapping

    def __get__(self, instance, owner):
        mapping = self.usermapping(instance)
        it = iter(mapping)
        try:
            k = next(it)
        except StopIteration:
            raise AmbiguousUserError("No users")
        try:
            next(it)
            raise AmbiguousUserError("{} users. Use the ``users`` property".format(len(mapping)))
        except StopIteration:
            return mapping[k]
        except AmbiguousUserError:
            if instance.name in mapping:
                return mapping[instance.name]
            raise


class Node(allegedb.graph.Node, rule.RuleFollower):
    """The fundamental graph component, which edges (in LiSE, "portals")
    go between.

    Every LiSE node is either a thing or a place. They share in common
    the abilities to follow rules; to be connected by portals; and to
    contain things.

    """
    __slots__ = ()
    engine = getatt('db')
    character = getatt('graph')
    name = getatt('node')
    no_unwrap = True

    def _get_rule_mapping(self):
        return RuleMapping(self)

    def _get_rulebook_name(self):
        try:
            return self.engine._nodes_rulebooks_cache.retrieve(
                self.character.name, self.name, *self.engine.btt()
            )
        except KeyError:
            return self.character.name, self.name

    def _get_rulebook(self):
        return rule.RuleBook(
            self.engine,
            self._get_rulebook_name()
        )

    def _set_rulebook_name(self, rulebook):
        character = self.character.name
        node = self.name
        cache = self.engine._nodes_rulebooks_cache
        try:
            if rulebook == cache.retrieve(character, node, *self.engine.btt()):
                return
        except KeyError:
            pass
        branch, turn, tick = self.engine.nbtt()
        cache.store(character, node, branch, turn, tick, rulebook)
        self.engine.query.set_node_rulebook(character, node, branch, turn, tick, rulebook)

    @property
    def portal(self):
        """Return a mapping of portals connecting this node to its neighbors."""
        return Dests(self)
    successor = adj = edge = portal

    @property
    def preportal(self):
        return Origs(self)
    predecessor = pred = preportal

    user = UserDescriptor()

    @property
    def users(self):
        return UserMapping(self)

    def __init__(self, character, name):
        """Store character and name, and initialize caches"""
        super().__init__(character, name)
        self.db = character.engine

    def __iter__(self):
        yield from super().__iter__()
        yield from self.extrakeys
        return

    def clear(self):
        for key in super().__iter__():
            del self[key]

    def __contains__(self, k):
        """Handle extra keys, then delegate."""
        if k in self.extrakeys:
            return True
        return super().__contains__(k)

    def __setitem__(self, k, v):
        super().__setitem__(k, v)
        self.send(self, key=k, val=v)

    def __delitem__(self, k):
        super().__delitem__(k)
        self.send(self, key=k, val=None)

    def portals(self):
        """Iterate over :class:`Portal` objects that lead away from me"""
        yield from self.portal.values()

    def successors(self):
        """Iterate over nodes with edges leading from here to there."""
        for port in self.portal.values():
            yield port.destination

    def preportals(self):
        """Iterate over :class:`Portal` objects that lead to me"""
        yield from self.preportal.values()

    def predecessors(self):
        """Iterate over nodes with edges leading here from there."""
        for port in self.preportal.values():
            yield port.origin

    def _sane_dest_name(self, dest):
        if isinstance(dest, Node):
            if dest.character != self.character:
                raise ValueError(
                    "{} not in {}".format(dest.name, self.character.name)
                )
            return dest.name
        else:
            if dest in self.character.node:
                return dest
            raise ValueError("{} not in {}".format(dest, self.character.name))

    def shortest_path_length(self, dest, weight=None):
        """Return the length of the path from me to ``dest``.

        Raise ``ValueError`` if ``dest`` is not a node in my character
        or the name of one.

        """

        return shortest_path_length(
            self.character, self.name, self._sane_dest_name(dest), weight
        )

    def shortest_path(self, dest, weight=None):
        """Return a list of node names leading from me to ``dest``.

        Raise ``ValueError`` if ``dest`` is not a node in my character
        or the name of one.

        """
        return shortest_path(
            self.character, self.name, self._sane_dest_name(dest), weight
        )

    def path_exists(self, dest, weight=None):
        """Return whether there is a path leading from me to ``dest``.

        With ``weight``, only consider edges that have a stat by the
        given name.

        Raise ``ValueError`` if ``dest`` is not a node in my character
        or the name of one.

        """
        try:
            return bool(self.shortest_path_length(dest, weight))
        except KeyError:
            return False

    @property
    def content(self):
        return NodeContent(self)

    def contents(self):
        return self.content.values()

    def delete(self):
        """Get rid of this, starting now.

        Apart from deleting the node, this also informs all its users
        that it doesn't exist and therefore can't be their avatar
        anymore.

        """
        if self.name in self.character.portal:
            del self.character.portal[self.name]
        if self.name in self.character.preportal:
            del self.character.preportal[self.name]
        for contained in list(self.contents()):
            contained.delete()
        for user in list(self.users.values()):
            user.del_avatar(self.character.name, self.name)
        branch, turn, tick = self.engine.nbtt()
        self.engine._nodes_cache.store(
            self.character.name, self.name,
            branch, turn, tick, False
        )
        self.engine.query.exist_node(
            self.character.name, self.name,
            branch, turn, tick, False
        )
        self.character.node.send(self.character.node, key=self.name, val=None)

    def one_way_portal(self, other, **stats):
        """Connect a portal from here to another node, and return it."""
        return self.character.new_portal(
            self, other, symmetrical=False, **stats
        )

    def one_way(self, other, **stats):
        """Connect a portal from here to another node, and return it."""
        return self.one_way_portal(other, **stats)

    def two_way_portal(self, other, **stats):
        """Connect these nodes with a two-way portal and return it."""
        return self.character.new_portal(
            self, other, symmetrical=True, **stats
        )

    def two_way(self, other, **stats):
        """Connect these nodes with a two-way portal and return it."""
        return self.two_way_portal(other, **stats)

    def new_thing(self, name, **stats):
        """Create a new thing, located here, and return it."""
        return self.character.new_thing(
            name, self.name, **stats
        )

    def historical(self, stat):
        """Return a reference to the values that a stat has had in the past.

        You can use the reference in comparisons to make a history
        query, and execute the query by calling it, or passing it to
        ``self.engine.ticks_when``.

        """
        return StatusAlias(
            entity=self,
            stat=stat
        )

    def __bool__(self):
        return self.name in self.character.node


# This file is part of LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""A dwarf hunting a kobold that hides in the bushes.

This script will initialize LiSEworld.db and the code libraries to run the
simulation described. To view it, run ELiDE from the same directory
as you ran this script from.

"""


import networkx as nx


def inittest(
        engine,
        mapsize=(10, 10),
        dwarf_pos=(0, 0),
        kobold_pos=(9, 9),
        shrubberies=20,
        dwarf_sight_radius=2,
        kobold_sprint_chance=.1
):
    # initialize world
    phys = engine.new_character('physical', data=nx.grid_2d_graph(*mapsize))
    kobold = phys.new_thing("kobold", kobold_pos)
    kobold['sprint_chance'] = kobold_sprint_chance
    kobold['_image_paths'] = ['atlas://rltiles/base.atlas/kobold_m']
    dwarf = phys.new_thing("dwarf", dwarf_pos)
    dwarf['sight_radius'] = dwarf_sight_radius
    dwarf['seen_kobold'] = False
    dwarf['_image_paths'] = ['atlas://rltiles/base.atlas/dwarf_m']
    # randomly place the shrubberies and add their locations to shrub_places
    n = 0
    locs = list(phys.place.keys())
    engine.shuffle(locs)
    shrub_places = []
    while n < shrubberies:
        loc = locs.pop()
        phys.add_thing(
            "shrub" + str(n),
            loc,
            cover=1,
            _image_paths=['atlas://rltiles/dc-mon.atlas/fungus'],
            _group='shrub'
        )
        shrub_places.append(loc)
        n += 1
    print('{} shrubberies: {}'.format(n, shrub_places))
    kobold['shrub_places'] = shrub_places

    # Basic day night cycle
    @engine.action
    def time_slipping(character, *, daylen: int, nightlen: int, twilightlen: float=0.0):
        if 'hour' not in character.stat:
            character.stat['hour'] = 0
            character.stat['day_period'] = 'dawn' if twilightlen else 'day'
            return
        twi_margin = twilightlen / 2
        hour = character.stat['hour'] = (character.stat['hour'] + 1) % (daylen + nightlen)
        if twilightlen:
            if hour < twi_margin or hour > daylen + nightlen - twi_margin:
                character.stat['day_period'] = 'dawn'
            elif twi_margin < hour < daylen - twi_margin:
                character.stat['day_period'] = 'day'
            elif daylen - twi_margin < hour < daylen + twi_margin:
                character.stat['day_period'] = 'dusk'
            else:
                character.stat['day_period'] = 'night'
        else:
            character.stat['day_period'] = 'day' if hour < daylen else 'night'


    # If the kobold is not in a shrubbery, it will try to get to one.
    # If it is, there's a chance it will try to get to another.
    @kobold.rule
    def shrubsprint(thing):
        shrub_places = sorted(list(thing['shrub_places']))
        if thing['location'] in shrub_places:
            shrub_places.remove(thing['location'])
            assert thing['location'] not in shrub_places
        whereto = thing.engine.choice(shrub_places)
        thing.travel_to(whereto)

    @shrubsprint.trigger
    def uncovered(thing):
        for shrub_candidate in thing.location.contents():
            if shrub_candidate.name[:5] == "shrub":
                return False
        thing.engine.info("kobold uncovered")
        return True

    @shrubsprint.trigger
    def breakcover(thing):
        if thing.engine.random() < thing['sprint_chance']:
            thing.engine.info("kobold breaking cover")
            return True

    @shrubsprint.prereq
    def not_traveling(thing):
        if thing.next_location is not None:
            thing.engine.info("kobold already travelling to {}".format(thing.next_location))
            return False
        else:
            return True

    @engine.method
    def set_kill_flag(eng):
        eng.character['physical'].thing['dwarf']['kill'] = True

    @dwarf.rule
    def fight(thing):
        method = thing.engine.method
        return "Kill kobold?", [("Kill", method.set_kill_flag), ("Spare", None)]

    @fight.trigger
    def sametile(thing):
        try:
            return (
                thing['location'] == thing.character.thing['kobold']['location']
            )
        except KeyError:
            return False

    @fight.prereq
    def kobold_alive(thing):
        return 'kobold' in thing.character.thing

    def aware(thing):
        # calculate the distance from dwarf to kobold
        from math import hypot
        try:
            bold = thing.character.thing['kobold']
        except KeyError:
            return False
        (dx, dy) = bold['location']
        (ox, oy) = thing['location']
        xdist = abs(dx - ox)
        ydist = abs(dy - oy)
        dist = hypot(xdist, ydist)
        # if it's <= the dwarf's sight radius, the dwarf is aware of the kobold
        return dist <= thing['sight_radius']

    fight.prereq(aware)

    @dwarf.rule
    def kill_kobold(thing):
        del thing.character.thing['kobold']
        del thing['kill']

    kill_kobold.prereq(kobold_alive)

    @kill_kobold.prereq
    def flag_set(thing):
        return 'kill' in thing

    @kill_kobold.trigger
    def unmerciful(thing):
        return thing['kill']

    @dwarf.rule
    def go2kobold(thing):
        thing.travel_to(thing.character.thing['kobold']['location'])

    go2kobold.trigger(aware)

    go2kobold.prereqs = ['kobold_alive']

    @go2kobold.prereq
    def kobold_not_here(thing):
        return 'kobold' not in thing.location.content

    @dwarf.rule
    def wander(thing):
        dests = sorted(list(thing.character.place.keys()))
        dests.remove(thing['location'])
        thing.travel_to(thing.engine.choice(dests))

    @wander.trigger
    def standing_still(thing):
        return thing.next_location is None


if __name__ == '__main__':
    from LiSE.engine import Engine
    from os import remove
    try:
        remove('LiSEworld.db')
    except FileNotFoundError:
        pass
    with Engine('LiSEworld.db', random_seed=69105) as engine:
        inittest(engine, shrubberies=20, kobold_sprint_chance=.9)
        engine.commit()
        print('shrub_places beginning: {}'.format(
            engine.character['physical'].thing['kobold']['shrub_places']
        ))
